/*
 *   ac.c --
 *
 *   This file contains the application control procedures. It is designed as 
 *   a stand alone file with no dependencies on the rest of nevit.
 *   When choosing to contol the sending source a hash table containing all
 *   the incoming streams with their information about the error and 
 *   transmission rates.
 *
 *  (C) 1995 Dorgham Sisalem, Bernd Deffner, GMD Fokus
 *
 *  See the copyright file for copyright information
 */

#include <stdio.h>
#include <stdlib.h>
#include <tcl.h>
#include <sys/types.h>
#include "types.h"
#include "ac.h"

extern int DEBUG;
static ac_t ac;
ac_t *ac_p=&ac;

static char rcsid[]  = "$Id: ac.c,v 1.4 1998/05/07 13:44:08 dor Exp $";

/*
 * add an entry to the table
 */
mem_t *ac_add_entry(ac_t *ac_p, u_int32 ssrc)
{
  int new;
  Tcl_HashEntry *entryPtr;
  mem_t *m=NULL;
  entryPtr = Tcl_CreateHashEntry(&ac_p->member_table, (char*)(long)ssrc, &new);
  m = (mem_t *)malloc(sizeof(mem_t));
  Tcl_SetHashValue(entryPtr, m);
  ac_p->members++;
  memset(m, 0, sizeof(mem_t));       /* initialize member structure */
  m->ssrc=ssrc;
  m->loss=0;
  m->filtered_loss=0;
  m->current_Q=0;
  m->max_Q=-1;
  m->TCP_Rate=0;
  m->time=0;
  m->min_rtt=100; /* set the initial minumum rtt to some very big value*/
  
  m->bbw=-1;
  return m;
}

/*
 * get the time in usecs
 */
u_int32 ac_time()
{
  struct timeval now;
  gettimeofday(&now,0);
  return((now.tv_sec*1000000)+now.tv_usec);
}



/*
 * convenience function for searching an entry in the hash table
 * returns pointer to member or NULL 
*/
mem_t* ac_find_entry(ac_t *ac_p, u_int32 ssrc)
{
  Tcl_HashEntry *entryPtr;
  
  entryPtr = Tcl_FindHashEntry(&ac_p->member_table, (char*)(long)ssrc);
  if (entryPtr == NULL)
    return NULL;
  return (mem_t *)Tcl_GetHashValue(entryPtr);
}

int ac_max_bandwidth()
{
 return ac_p->max_bandwidth;
}

/* 
 * deletes an entry from the hash table and frees the allocated 
 * memory for the member entry
*/
void ac_delete_entry(u_int32 ssrc)
{
  Tcl_HashEntry *entryPtr;
  mem_t *m=NULL;
  Tcl_HashSearch search;
  if((int)(ssrc)>0) {
    entryPtr = Tcl_FindHashEntry(&ac_p->member_table, (char*)(long unsigned )ssrc);
    if (entryPtr != NULL) {
      m = (mem_t *)Tcl_GetHashValue(entryPtr);
      Tcl_DeleteHashEntry(entryPtr);
      ac_p->members--;
      free ((char *)m);
      m=NULL;
    }
    return;
  }
  /* if ssrc<0 delete all members */
  if(&ac_p->member_table) {
    for (entryPtr = Tcl_FirstHashEntry(&ac_p->member_table, &search); 
	 entryPtr != NULL;
	 entryPtr = Tcl_NextHashEntry(&search)) {
      m = (mem_t *) Tcl_GetHashValue(entryPtr);
      Tcl_DeleteHashEntry(entryPtr);
      ac_p->members--;
      free ((char *)m);
      m=NULL;
    }
  }
}


/*
*  Initialisation of all ac related parameters
*/  

void ac_init_all(u_int32 ssrc, int algorithm, double alpha, int max_bandwidth,
		 double multiplicative_decrease,
		 double additive_increase,
		 double loss_lower_threshold,
		 double loss_upper_threshold,
		 double congested_recv_fraction,
		 double loaded_recv_fraction)
{
  
  Tcl_InitHashTable(&ac_p->member_table, TCL_ONE_WORD_KEYS);
  ac_p->control			= 1;
  ac_p->algorithm		= algorithm;
  ac_p->alpha			= alpha;
  ac_p->max_bandwidth		= 0.9*max_bandwidth;
  ac_p->min_bandwidth		= 0;
  ac_p->multiplicative_decrease	= multiplicative_decrease/100.0;
  ac_p->additive_increase	= additive_increase;
  ac_p->loss_lower_threshold	= loss_lower_threshold/100.0;
  ac_p->loss_upper_threshold	= loss_upper_threshold/100.0;
  ac_p->congested_recv_fraction	= congested_recv_fraction/100.0;
  ac_p->loaded_recv_fraction	= loaded_recv_fraction/100.0; 
  ac_p->members			= 1;
  ac_p->ssrc			= ssrc;
  ac_p->time                    = 0;
  ac_p->aif                     = additive_increase;
}

/* just init the ssrc and member table */
void ac_init(u_int32 ssrc)
{
  
  Tcl_InitHashTable(&ac_p->member_table, TCL_ONE_WORD_KEYS);
  ac_p->control			= 1;
  ac_p->members			= 1;
  ac_p->ssrc			= ssrc;
  ac_p->time                    = 0;
}

/* 
 * update the control information of the sending source
 */
 
void ac_update(int tag,int parameter)
{
  switch ( tag ) {
  case CONTROL: 
    ac_p->control		= parameter;
    break;
  case ALGORITHM:
    ac_p->algorithm		= parameter; 
    break;
  case ALPHA:
    ac_p->alpha			= parameter;
    break;
  case MAX_BANDWIDTH:
    ac_p->max_bandwidth		= 0.9*parameter;
    break;
  case MIN_BANDWIDTH:
    ac_p->min_bandwidth		= parameter;
    break;
  case MULTIPLICATIVE_DECREASE:
    ac_p->multiplicative_decrease	= parameter/100.0;
    break;
  case ADDITATIVE_DECREASE:
    ac_p->additive_increase	= parameter;
    break;
  case LOSS_LOWER_THRESHOLD:
    ac_p->loss_lower_threshold	= parameter/100.0;
    break;
  case LOSS_UPPER_THRESHOLD:
    ac_p->loss_upper_threshold	= parameter/100.0;
    break;
  case CONGESTED_RECV_FRACTION:
    ac_p->congested_recv_fraction	= parameter/100.0;
    break;
  case LOADED_RECV_FRACTION:
    ac_p->loaded_recv_fraction	= parameter/100.0;
    break;
  default :
    perror(" ac_update: Unknown parameter ");
  }
}

/*
 * This function analyses the information contained in the RTP database 
 * and returns the actual network state.
 * The network state is set according to the maximum loss rate of all
 * receivers.
 */
network_state_t worst_receiver(ac_t *ac_p)
{
  float max_loss;
  mem_t *m;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  network_state_t state=UNLOADED;

  max_loss = -1;

  /*
   *  search through the hash table and look for the highest 
   *  filtered loss value
   */
  for (entryPtr = Tcl_FirstHashEntry(&ac_p->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    if (m->ssrc == ac_p->ssrc) continue;  /* neglect entry for sender */
    if (m->filtered_loss > max_loss)
      max_loss = m->filtered_loss;
  }

  /*
   *  linear regulator with deadzone
   */
  if (max_loss >= ac_p->loss_upper_threshold)
    state = CONGESTED;
  else if (max_loss >= ac_p->loss_lower_threshold) 
    state = LOADED;
  else
    state = UNLOADED;
  return state;
}


/* 
 * Count the number of receivers that are congested, loaded and unloaded
 * and determine network state 
 */
network_state_t congestion_ratio(ac_t *ac_p)
{
  static network_state_t old_state = LOADED;
  network_state_t state;
  int unloaded, loaded, congested;
  mem_t *m=NULL;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  float v;

  unloaded = 0;
  loaded = 0;
  congested = 0;

  /*
   *  search through the hash table and determine the number of 
   *  receivers that are congested, loaded or unloaded
   */
  for (entryPtr = Tcl_FirstHashEntry(&ac_p->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    printf("myself= %lu recv=%ui\n", (unsigned long int)ac_p->ssrc, 
	   m->ssrc);
    if (m->ssrc == ac_p->ssrc) continue;    /* ignore entry for sender */
    v = m->filtered_loss;
    printf("loss of recv %x is %f\n", m->ssrc, v);
    if (v >= ac_p->loss_upper_threshold)
      congested++;
    else if (v >= ac_p->loss_lower_threshold) 
      loaded++;
    else 
      unloaded++;
  }

  if ( (double)congested / ac_p->members - 1
      >= ac_p->congested_recv_fraction)
    state = CONGESTED;
  else if ((double)loaded / ac_p->members - 1
	   >= ac_p->loaded_recv_fraction)
    state = LOADED;
  else 
    state = UNLOADED;
  old_state = state;

  if(DEBUG)
    fprintf(stderr, "receivers=%d #unloaded=%d #loaded=%d #congested=%d"
	    " state=%d", ac_p->members - 1, unloaded, loaded, congested,state);
  
  return state;
}

/*
 * low-pass filter to compute the short time statistics                 
 * the smaller a, the greater is the influence of the old value 
 */
float ac_filter(float old, float new, float a)
{
  if(DEBUG)
    fprintf(stderr,"old=%f new=%f, a=%f result=%f\n",old,new,a,(1-a)*old + a*new);
  return ( (1-a)*old + a*new );
}


/*
 * This function interpretes the network state and adjusts the bandwidth
 * of the video agent accordingly
 */
int adjust(ac_t *ac_p, network_state_t state,int bandwidth) 
{
  /*
   * Compute new bandwidth
   */
  float current_bw=(float) bandwidth;
  if (ac_p->control) {
    switch(state) {
    case UNLOADED:    current_bw= current_bw+ ac_p->additive_increase;
      break;
    case LOADED:      break;
    case CONGESTED:   current_bw = (current_bw*ac_p->multiplicative_decrease);
    }
  }
  if (current_bw <=ac_p->min_bandwidth)
    current_bw = ac_p->min_bandwidth;
  else if (current_bw >= ac_p->max_bandwidth)
    current_bw= ac_p->max_bandwidth;

  return (int)current_bw;
}


/* the direct_adjustment algorithm reduces the sending bandwidth directly 
 * to the level reported in the receiver reports. the reduction is then
 * maintained untill a new RR arrives from the same receiver indicating that
 * the loss is beneath the lower loss threshold.
 * In the case that after reducing the rate another RR message was received
 * from another receiver with a heigher loss ratio then the rate is only
 * reduced by the difference of the losses reported in this RR and the one
 * that caused the original loss. Afterwards we have to wait for another RR
 * from this receiver to start increasing the rate again.
 */ 
 
int direct_adjustment(int ssrc, double loss, int bandwidth)
{
  int bw=bandwidth;
  int static result=UNLOADED;
  if(DEBUG)
    fprintf(stderr," bw=%i ac_p-loss=%f loss=%f ssrc=%i, ac_ssrc=%i\n",bandwidth, ac_p->loss,loss,ssrc,ac_p->ssrc_loss);

  if(ac_p->ssrc_loss==ssrc){ /* RR from the ssrc with the heighest loss */ 
    if(loss>ac_p->loss_lower_threshold){
      bw=bandwidth*(1-loss+ac_p->loss_lower_threshold);
      result=CONGESTED;
    }
    else{
      result=UNLOADED;
    }
    ac_p->loss=loss;
  } else {
    if(loss>ac_p->loss){/* a new loss reported*/
      bw=bandwidth*(1-loss+ac_p->loss+ac_p->loss_lower_threshold);
      ac_p->loss=loss;
      ac_p->ssrc_loss=ssrc;
      result=CONGESTED;
    }
  }

  if(result==UNLOADED) {
    bw+=ac_p->additive_increase;
  }

  /* check if the bandwidth is still in the desired limits */
 /* if (bw <=ac_p->min_bandwidth)
     bw = ac_p->min_bandwidth;*/
  else if (bw >= ac_p->max_bandwidth)
    bw= ac_p->max_bandwidth;
  return bw;
}

/*
 * update the entries calculation for this member
 */
void ac_update_entry(mem_t* m )
{
  /* determine the minumum seen round trip time and consider it as the 
     propagation delay */
  if(m->min_rtt> m->rtt)
    m->min_rtt=m->rtt;

  if(DEBUG)
    fprintf(stderr," ac_update min_rtt=%f ",m->min_rtt);
  

  /* determine the current queuing delay */
  m->current_Q = m->rtt - m->min_rtt;

  /* set the current update time */
  m->time=ac_time();
  
  if(DEBUG)
    fprintf(stderr," current_Q=%f max_Q=%f time=%u \n", m->current_Q, m->max_Q, m->time);
}

/*
 * calculate for each member the bandwdith it could have been getting
 * under the condition of current bandwdith and the measured delay and loss
 */
double ac_get_LDBA_bw(ac_t* ac_p,mem_t* m,double current_bandwidth) 
{
  double bw=current_bandwidth;
  int min_bbw=ac_min_bbw(ac_p);
  
  if(min_bbw<1)
    min_bbw=ac_p->max_bandwidth;
  if(DEBUG)
    fprintf(stderr," LDBA ssrc=%i  bw=%f loss=%f MD=%f AIF=%f \n ----min_bbw=%i\n",m->ssrc,bw, m->loss,ac_p->multiplicative_decrease*100, ac_p->additive_increase,min_bbw);
  
  if(m->loss > ac_p->loss_lower_threshold) {
    if(m->loss > ac_p->loss_upper_threshold) {
      bw*=0.5;
    } else {
      bw-=bw*m->loss*(ac_p->multiplicative_decrease*100);
    }
    m->max_Q=m->current_Q;
    if(DEBUG)
      fprintf(stderr," loss=%f bw=%f \n",m->loss, bw);
    return (bw);
  }
  /* losses lower than lower threshold */
  {
    double delay;
    double window;
    double aif;
   
    /*delay=m->min_rtt+m->max_Q;*/
    delay=m->rtt;

    /* calculate the TCP equivalent increase during UPDATE_INTERVAL and no 
     * losses and rtt of delay in kb/s*/
    window=1.024*(UPDATE_INTERVAL/1000000.0)*8/(delay*2);

    
    /* are we under the lower loss threshold but still have some losses? */
    if(m->loss>0.00001){
      m->TCP_Rate=(1.22*8*1.024)/(delay*sqrt(m->loss));
      if(DEBUG)
	fprintf(stderr," delay=%f loss=%f TCP_Rate=%f \n",delay,m->loss,m->TCP_Rate);

      /* make sure TCP_Rate has an acceptable value */
      if(m->TCP_Rate>ac_p->max_bandwidth||m->TCP_Rate<0)
	m->TCP_Rate=ac_p->max_bandwidth;
      
      if(bw>m->TCP_Rate) {
	bw=bw-(bw*m->loss*ac_p->multiplicative_decrease*100);
	return(bw);
      }
    }
    { /* no losses */
      /* set the additive factor in accordance with the highest bandwdith */
      double aif_factor=1 - (double)(current_bandwidth/ac_p->max_bandwidth);
      
      /*make sure the factors have maximally the value 1 and a minimum of 0*/
      aif_factor= ((aif_factor<1) ? aif_factor :1);
      aif_factor= ((aif_factor<0) ? 0: aif_factor );
      
      aif=(ac_p->aif*aif_factor)+ac_p->aif;
      if( aif<ac_p->additive_increase)
	aif=ac_p->additive_increase;
      if(aif>window)
	aif=window;
      
      bw+=aif;
      if(bw>ac_p->max_bandwidth)
	bw=ac_p->max_bandwidth;
      if(DEBUG)
	fprintf(stderr,"window=%f aif=%f  bw=%f aif_factor=%f rtt=%f\n",window,aif,bw,aif_factor,m->rtt);
    }
    return (bw);
  }
}
    

   
/*
 * find the smallest possible bandwdith to use 
 */
double ac_get_min_entry(ac_t *ac_p,double current_bandwidth)
{
  mem_t *m;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  double bandwidth= ac_p->max_bandwidth;

  /*
   *  search through the hash table and look for the lowest
   *  bandwidth
   */
  double tmp_bw;
  for (entryPtr = Tcl_FirstHashEntry(&ac_p->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    if (m->ssrc == ac_p->ssrc) continue;  /* neglect entry for sender */
    tmp_bw=ac_get_LDBA_bw(ac_p,m,current_bandwidth);
    bandwidth= (bandwidth > tmp_bw) 
      ?tmp_bw : bandwidth; 
  }
  if (bandwidth <=ac_p->min_bandwidth)
    bandwidth = ac_p->min_bandwidth;
  else if (bandwidth >= ac_p->max_bandwidth)
    bandwidth= ac_p->max_bandwidth;

  if(bandwidth > current_bandwidth) {
    ac_p->aif=bandwidth-current_bandwidth;
  } else {
    ac_p->aif=ac_p->additive_increase;
  }
  if(DEBUG)
    fprintf(stderr," current_bw=%f, bandwdith=%f  aif=%f \n",current_bandwidth,bandwidth,ac_p->aif);

  return bandwidth;
}

/* 
 * update the value of the bottleneck bandwidth for member ssrc
 */
void ac_update_bbw(int ssrc, double bbw)
{
  mem_t *m;
  m = ac_find_entry(ac_p, ssrc);
  if (m == NULL) {
    m = ac_add_entry(ac_p, ssrc);
  }
  m->bbw=bbw/1000; /* convert into kb/s */
}

int ac_min_bbw(ac_t *ac_p)
{
  mem_t *m;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  static double min_bbw=0;
  int first=1;
  

  /*
   *  search through the hash table and look for the smallest
   *  bbw value
   */
  for (entryPtr = Tcl_FirstHashEntry(&ac_p->member_table, &search); 
       entryPtr != NULL;
       entryPtr = Tcl_NextHashEntry(&search)) {
    m = (mem_t *) Tcl_GetHashValue(entryPtr);
    if (m->ssrc == ac_p->ssrc) continue;  /* neglect entry for sender */
    if(m->bbw>0) {
      if(first) {
	first=0;
	min_bbw=m->bbw;
      } else {
	if(m->bbw<min_bbw)
	  min_bbw=m->bbw;
      }
    }
    /* reset bbw to -1 */
    m->bbw=-1;
  }
  if(DEBUG)
     fprintf(stderr," ac_min min_bbw=%f \n",min_bbw);
  return(min_bbw);
}


/* 
 * call the selected algorithm
 */
int ac_qos_control(int ssrc, double loss, double rtt,int bandwidth)
{
  network_state_t result=UNLOADED;
  mem_t *m=NULL;

  if(ac_p->algorithm==LOSS_BASED_ADJUSTMENT){
    return direct_adjustment(ssrc,loss,bandwidth);
  }
  m = ac_find_entry(ac_p, ssrc);
  if (m == NULL) {
    m = ac_add_entry(ac_p, ssrc);
  }
  /*m->loss = loss;*/
  m->loss=ac_filter(m->loss, loss,ac_p->alpha);
  m->rtt=rtt;
  
  if(ac_p->algorithm==LOSS_DELAY_BASED_ADJUSTMENT){
    ac_update_entry(m);
    
    /* check if it is time to do an update */
    if(ac_time()> ac_p->time) 
      ac_p->time=ac_time()+UPDATE_INTERVAL; /* set the time for the next 
					       update interval */
    return(ac_get_min_entry(ac_p,bandwidth)); 
  }

  switch(ac_p->algorithm) {
    case WORST_RECEIVER: 
      result = worst_receiver(ac_p);
      break;
    case CONGESTION_RATIO:
      result = congestion_ratio(ac_p);
      break;
  }
  return adjust(ac_p,result,bandwidth);
}
