/*
 * color_map.c  ---
 *
 * procedures for allocating a private color map for
 * pseudo color environments
 *
 * The code is partially taken from the vic video tool
 *
 * 1996 Dorgham Sisalem GMD Fokus
 *
 * The code presented here is ported into C by the author from
 * the code developed by the University of California in C++. The code
 * is part of the vic video tool developed by the University.
 *
 * Copyright (c) 1993-1995 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Dorgham Sisalem
 *       GMD Fokus
 * 4. Neither the name of the GMD Fokus nor of the author may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "color_map.h"
#include "color_cube.h"
#include "bsd-endian.h"
#include <math.h>


static char rcsid[]  = "$Id: color_map.c,v 1.3 1998/09/01 15:38:17 dor Exp $";

#define LIM(x) ((x) >= 255 ? 255 : ((x) < 0 ? 0 : (x)))

double gamma_=0.7;

#define CDIST(x,y,z)  ((x)*(x) + (y)*(y) + (z)*(z))
/************* Local procedures declaration ************************/
int alloc_color(int r, int g, int b);
int alloc_colors_pseudo();
int alloc_grays();


/**************************** local variables ************************/
static int alloc_priv=0; /* are we using a private color map or not */


/**************************************************************************/
unsigned long  stdfreecols[256];
/*
 * free the allocated colors
 */
void free_colors()
{
	if (ncolor_ > 0) {
		XFreeColors(dpy_, colormap_, pixel_, ncolor_, 0);
		ncolor_ = 0;
	}
}

/**************************************************************************/
/* 
 * alloc a private color map for pseudo color environments 
 */
int my_alloc_color()
{
	colormap_=XCreateColormap(dpy_,win_,  visual_, AllocNone);
	if (!colormap_) {
	  perror("Couldn't create private colormap!\n");
	  return (-1);
	}
	if(alloc_grays() < 0 ||alloc_colors_pseudo()<0) {
	    free_colors();
	    return(-1);
	}
	return(1);
}


/**************************************************************************/
int alloc_color(int r, int g, int b)
{
        int pixel;
	XColor xc;
	if (r == g && r == b && (r & 7) == 0 && r > 0) {
		/*
		 * This color is one level away from the gray ramp
		 * used in nv, ghostscript, and the LBL-modified
		 * tk library.  Change it so we'll end up sharing
		 * the colormap entry.  The error won't be perceptible.
		 */
		 r -= 1;
		 g -= 1;
		 b -= 1;
	 }
	xc.red = r << 8;
	xc.green = g << 8;
	xc.blue = b << 8;
	xc.pixel=0;
	xc.flags=0;
	xc.pad=0;
	if (XAllocColor(dpy_,colormap_ , &xc) == 0) {
	  fprintf(stderr," Could not XAllocColor \n");
	  free_colors();
	  return (-1);
	}
	pixel = xc.pixel;
	{ 
	  color *p = &(colors_[pixel]);
	  double r,g,b;
	  pixel_[ncolor_] = pixel;
	  if (++ncolor_ >= 256) {
		fprintf(stderr, "nevit: colormap overflow (internal error)\n");
		return(-1);
	  }
	  p->r = xc.red >> 8;
	  p->g = xc.green >> 8;
	  p->b = xc.blue >> 8;
	  

	  r = p->r;
	  g = p->g;
	  b = p->b;

	  p->y = (short)(0.299 * r + 0.587 * g + 0.114 * b + 0.5);
	  p->u = (short)(-0.1687 * r - 0.3313 * g + 0.5 * b + 0.5) + 128;
	  p->v = (short)(0.5 * r - 0.4187 * g - 0.0813 * b + 0.5) + 128;
	}
	return (pixel);
}

/**************************************************************************/
int allocate(int ncolor, const u_char* cmap, u_char* lut)
{
	int LUT[256];
	int i,pixel;
	for (i = 0; i < ncolor; ++i) {
		int r = cmap[0];
		int g = cmap[1];
		int b = cmap[2];
#ifdef notdef
		if (r < 240 && g < 240 && b < 240) {
			r += 16;
			g += 16;
			b += 16;
		}
#endif
		pixel = alloc_color(r, g, b);
		if (pixel < 0) 
			return (-1);
		cmap += 3;
		LUT[i] = pixel;
	}
	/*
	 * Map from colormap indices to allocated pixels.
	 */
	for (i = 0; i < 65536; ++i)
		lut[i] = LUT[lut[i]];

	return (0);
}

/**************************************************************************/
int installLUT(int def_ncolor, const u_char* def_cmap,
			      const u_char* def_lut)
{
	u_char cmap[3*256];
	int ncolor = -1;
	if (lut_){ free(lut_); lut_=NULL;}
	lut_ = malloc(sizeof(u_char)*65536);
	if (ncolor < 0) {
		ncolor = def_ncolor;
		memcpy(cmap, def_cmap, 3 * ncolor);
		memcpy(lut_, def_lut, 65536);
	}
	if (allocate(ncolor, cmap, (u_char *)lut_) < 0) {
		free(lut_);
		lut_ = 0;
		return (-1);
	}
	return (0);
}

/**************************************************************************/
int alloc_colors_pseudo()
{
     u_char *lut, *lut_tmp;
     int i,k,pixel;
	if (installLUT(cube_ncolor, cube_cmap, cube_lut) < 0)
		return (-1);

	/*
	 * Fold ordered dither matrix into lookup table.
	 */
	lut = (u_char *)lut_;
	lut_tmp=NULL;
	if((lut_tmp = malloc(sizeof( u_char)*16 * 65536))==NULL){
	  perror("can not allocate mem ");
	  return(-1);
	}
	for (i = 0; i < 65536; ++i) {
		int u0 = (i >> 11) << 3;
		int v0 = ((i >> 6) & 0x1f) << 3;
		int y0 = (i & 0x3f) << 2;
			
		for (k = 0; k < 16 ; ++k) {
			int y = y0 + 2 * tessellate[k].y;
			int u = u0 + 2 * tessellate[k].u;
			int v = v0 + 2 * tessellate[k].v;

			u = LIM(u);
			v = LIM(v);
			y = LIM(y);

			/* convert to right precision (5U:5V:6Y) */
			y >>= 2;
			u >>= 3;
			v >>= 3;

			/* lookup best match */
			pixel = lut[(u << 11) | (v << 6) | y];
			lut_tmp[(i << 4) | k] = pixel;
		}
	}
	lut_=(unsigned long *)lut_tmp;
	free(lut);

	return (0);
}

/**************************************************************************/
int alloc_grays()
{
#define DELTA (256 / 32)
  	int pixel[32];
	int i;
	for (i = 0; i < 32; ++i) {
		int luma = (i + 1) * DELTA - 1;
		color c;
		c.r = c.g = c.b = luma;
		pixel[i] = alloc_color(c.r,c.g,c.b);
		if (pixel[i] < 0)
			return (-1);
	}
	for (i = 0; i < 256; ++i) {
		int level = (i + 1) / DELTA;
		int err;
		if (level == 0)
			err = 0;
		else {
			level -= 1;
			err = (i + 1) & (DELTA - 1);
		}
		grey_lut_[i] = pixel[level] | err << 8;
	}
	for (; i < 512; ++i)
		grey_lut_[i] = pixel[31];
		
	return (0);
}



/**************************************************************************/
int allocate_colormap(  Display *theDisp, Visual* theVisual, Window win,Colormap theCmap)
{
  dpy_=theDisp;
  visual_=theVisual;
  win_=win;
  colormap_=theCmap;
  ncolor_=0;

  if(alloc_grays() < 0 || alloc_colors_pseudo()<0) {
    fprintf(stderr,"ran out of colors: reverting to private color map\n");
    free_colors();
    if(my_alloc_color()<0) {
      fprintf(stderr," could not allocate private color map\n");
      return (-1);
    }
    alloc_priv=1;
    XSetWindowColormap(dpy_,win_,colormap_);
    /*XInstallColormap(dpy_,colormap_); */
    XSync(dpy_,False);
  }
  return(1);
}

void cm_installcolor()
{
  if(alloc_priv) {
    XInstallColormap(dpy_,colormap_); 
    XSync(dpy_,False);
  }
}

void cm_uninstallcolor()
{
  if(alloc_priv) {
    XUninstallColormap(dpy_,colormap_); 
    XSync(dpy_,False);
  }
}
  
/************************** End color alloc ******************************/


