/*
 * rtp.h  --  RTP header file
 * RTP draft: November 1994 version
 *
 *
 * Constants and structures based on the 12/15/92 draft of the RTP protocol
 * Internet Draft. This information is still subject to change.
 *
 * $Id: rtp.h,v 1.4 1998/05/07 13:35:21 dor Exp $
 *
 *  See the copyright file for copyright information
 */

#ifndef _rtp_h_
#define _rtp_h_

#define RTP_SEQ_MOD (1<<16)
#define RTP_TS_MOD  (0xffffffff)

/* RTCP interval computation */

#define SESSION_BANDWIDTH 1000000. / 8.
#define TARGET_RTCP_BANDWIDTH 0.05 * SESSION_BANDWIDTH

/*
 * Gain (smoothing constant) for the low-pass filter that
 * estimates the average RTCP packet size
 */
#define RTCP_SIZE_GAIN (1./16.)

/*
* Current type value.
*/
#define RTP_VERSION 2

#define RTP_MAX_SDES 256   /* maximum text length for SDES */

typedef enum {
  RTCP_SR   = 200,
  RTCP_RR   = 201,
  RTCP_SDES = 202,
  RTCP_BYE  = 203,
  RTCP_APP  = 204
} rtcp_type_t;

typedef enum {
  RTCP_SDES_END    =  0,
  RTCP_SDES_CNAME  =  1,
  RTCP_SDES_NAME   =  2,
  RTCP_SDES_EMAIL  =  3,
  RTCP_SDES_PHONE  =  4,
  RTCP_SDES_LOC    =  5,
  RTCP_SDES_TOOL   =  6,
  RTCP_SDES_NOTE   =  7,
  RTCP_SDES_PRIV   =  8, 
  RTCP_SDES_IMG    =  9,
  RTCP_SDES_DOOR   = 10,
  RTCP_SDES_SOURCE = 11
} rtcp_sdes_type_t;

#if PC
typedef struct {
  unsigned int cc:4;       /* CSRC count */
  unsigned int x:1;        /* header extension flag */
  unsigned int p:1;        /* padding flag */
  unsigned int version:2;  /* protocol version */
  unsigned int pt:7;       /* payload type */
  unsigned int m:1;        /* marker bit */
  u_int16 seq;             /* sequence number */
  u_int32 ts;              /* timestamp */
  u_int32 ssrc;            /* synchronization source */
} rtp_hdr_t;
#else
typedef struct {
  unsigned int version:2;  /* protocol version */
  unsigned int p:1;        /* padding flag */
  unsigned int x:1;        /* header extension flag */
  unsigned int cc:4;       /* CSRC count */
  unsigned int m:1;        /* marker bit */
  unsigned int pt:7;       /* payload type */
  u_int16 seq;             /* sequence number */
  u_int32 ts;              /* timestamp */
  u_int32 ssrc;            /* synchronization source */
} rtp_hdr_t;
#endif

#if PC
typedef struct {
  unsigned int count:5;    /* varies by payload type */
  unsigned int p:1;        /* padding flag */
  unsigned int version:2;  /* protocol version */
  unsigned int pt:8;       /* payload type */
  u_int16 length;          /* packet length in words, without this word */
} rtcp_common_t;
#else
typedef struct {
  unsigned int version:2;  /* protocol version */
  unsigned int p:1;        /* padding flag */
  unsigned int count:5;    /* varies by payload type */
  unsigned int pt:8;       /* payload type */
  u_int16 length;          /* packet length in words, without this word */
} rtcp_common_t;
#endif
/* reception report */
typedef struct {
  u_int32 ssrc;            /* data source being reported */
  unsigned int fraction:8; /* fraction lost since last SR/RR */
  int lost:24;             /* cumulative number of packets lost (signed!) */
  u_int32 last_seq;        /* extended last sequence number received */
  u_int32 jitter;          /* interarrival jitter */
  u_int32 lsr;             /* last SR packet from this source */
  u_int32 dlsr;            /* delay since last SR packet */
} rtcp_rr_t;

typedef struct {
  u_int8 type;             /* type of SDES item (rtcp_sdes_type_t) */
  u_int8 length;           /* length of SDES item (in octets) */
  char data[1];            /* text, not zero-terminated */
} rtcp_sdes_item_t;

/* one app packet from the receiver side */
typedef struct {
  u_int32 ssrc; /* source number of the connection being reported on */
  u_int32 bbw;  /* estimated bandwidth */
} rtcp_app_rec_t;

/* one app packet from the sender side */
typedef struct {
  u_int32 seq; /*sequence number of measurement packet */
  rtcp_app_rec_t app_rec[1];
} rtcp_app_send_t;

/* one RTCP packet */
typedef struct {
  rtcp_common_t common;    /* common header */
  union {
    /* sender report (SR) */
    struct {
      u_int32 ssrc;        /* source this RTCP packet refers to */
      u_int32 ntp_sec;     /* NTP timestamp */
      u_int32 ntp_frac;
      u_int32 rtp_ts;      /* RTP timestamp */
      u_int32 psent;       /* packets sent */
      u_int32 osent;       /* octets sent */ 
      /* variable-length list */
      rtcp_rr_t rr[1];
    } sr;

    /* reception report (RR) */
    struct {
      u_int32 ssrc;        /* source this generating this report */
      /* variable-length list */
      rtcp_rr_t rr[1];
    } rr;

    /* BYE */
    struct {
      u_int32 src[1];      /* list of sources */
      u_int8 length;       /* length of reason text */
      /* can't express trailing text */
    } bye;

    /* source description (SDES) */
    struct rtcp_sdes_t {
      u_int32 src;              /* first SSRC/CSRC */
      rtcp_sdes_item_t item[1]; /* list of SDES items */
    } sdes;

    /* application specific packet (for measuring bottleneck bandwdith)*/
    struct rtcp_app_t {
      u_int32 src;              /* SRC */      
      char name[4];    
      union {
	rtcp_app_rec_t app_rec[1];
	rtcp_app_send_t app_send;
      } r;
    } app;          
  } r;
} rtcp_t;


/* RTP standard content encodings for video */
#define RTP_PT_YUV              22      /* YUV */
#define RTP_PT_SCR              23      /* simple conditional replenishment */
#define RTP_PT_MPEG             24      /* MPEG II? */
#define RTP_PT_CELLB            25      /* Sun CellB */
#define RTP_PT_JPEG             26      /* JPEG */
#define RTP_PT_CUSEEME          27      /* Cornell CU-SeeMe */
#define RTP_PT_NV               28      /* Xerox PARC nv */
#define RTP_PT_DVC              29      /* BBN dvc */
#define RTP_PT_BOLT             30      /* Bolter */
#define RTP_PT_H261             31      /* CCITT H.261 */
#define RTP_PT_GIF              50      /* GIF */

/* self-defined, be careful */

/*
 * Motion JPEG encapsulation.
 *
 * 0                   1                   2                   3
 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |      MBZ      |                frag offset                    |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |      Type     |       Q       |     Width     |     Height    |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 
 * Type = Index into a table of predefined JPEG parameters
 * Width = Width of input in 8-pixel blocks
 * Height = Height of input in 8-pixel blocks
 * Q = Quality factor (0-100 = std, >100 = custom)
 * Frag offset = The byte offset into the frame for the data in 
 * this packet
 */
typedef struct {
        unsigned int mbz:8;
	unsigned int off:24;    /* fragment offset */
        u_int8 type;            /* id of jpeg decoder params */
        u_int8 q;               /* quantization factor (or table id) */
        u_int8 width;           /* 1/8 frame width */
        u_int8 height;          /* 1/8 frame height */
} jpeg_hdr_t;

/*
 * GIF encapsulation.
 *
 * 0                   1                   2                   3
 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |                     frag offset                               |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 
 * 
 * Frag offset = The byte offset into the frame for the data in 
 * this packet
 */
typedef struct {
  u_int32 off;            /* fragment offset */
  u_int16 size;           /* size of gif file */
  u_int16 id  ;           /* id of gif file */
} gif_hdr_t;


/* 
 * CELLB encapsulation
 *
 * The packetization of the CellB bytestream is designed to make the
 * resulting packet stream robust to packet loss.  To achieve this goal, an
 * additional header is added to each RTP packet to uniquely identify the
 * location of the first cell of the packet within the current frame.  In
 * addition, the width and height of the frame in pixels is carried in each
 * CellB packet header.  Although the size can only change between frames,
 * it is carried in every packet to simplify the packet encoding.
 * 
 *  0                   1                   2                   3
 *  0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |       Cell X Location         |      Cell Y Location          |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |       Width of Image          |      Height of Image          |
 * +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
 * |                     Compressed CellB Data                     |
 * |                             ....                              |
 * +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
 */
typedef struct {
  u_int16 x, y;
  u_int16 w,h;
} cellb_hdr_t;



/*
 * YUV encapsulation 
 *
 * only width and height of the compressed picture are needed
*/
typedef struct {
  u_int16 width, height;
} yuv_hdr_t;


#endif
