/*
alpha_audio -- DEC AudioFile interface for Nevot
written by Philipp Harrschar
July 1995

$Id: alpha_audio.c,v 1.1.1.1 1997/11/12 17:16:26 dor Exp $
*/

#include <fcntl.h>       /* O_RDWR */

#include "audio_generic.h"
#include "ahw.h"
#include "notify.h"
#include <sys/time.h> 

/* DEC AudioFile */
#include <AF/AFlib.h>
#include <AF/AFUtils.h>

/* for debugging */
#define AHWDBG 0
#define ahw_debug(level, x) if (AHWDBG >= level) fprintf(stderr, x)
#define ahw_debug_arg1(level, x,y) if (AHWDBG >= level) fprintf(stderr, x, y)
#define ahw_debug_arg2(level, x,y,z) if (AHWDBG >= level) fprintf(stderr, x, y, z)
#define ahw_debug_arg3(level, x,y,z,a) if (AHWDBG >= level) fprintf(stderr, x, y, z, a)

/* connection to audiofile server */
static AFAudioConn *conptr = NULL; /* connection to server*/
static char *con_nameptr = ""; /* connection name */

/* attributes for contexts */
static AFSetACAttributes attributes;

/* read and write context, values do not differ */
AC ac = NULL;

/* device type, support for baseboard device and Aj300 (lin16) */
/* j300 server is determined by device encoding lin16 */
static enum device_t {D_BASE, D_J300}; 
static int device_type = D_BASE;

/* af device number to use */
static af_device = 0;

/* maximum AFGetTime retval */
#define AFMAXTIME 4294967296

/* first packet to read after open  */
static int first_read = 1;
/* first call of ahw_open */
static int first_open = 1;
/* first call of ahw_write */
static int first_write = 1;

int ahw_changed = 0;

/* time scope for reading bytes */
static ATime af_read_start,af_read_end; 
static int delta;

/* delay statistics between ahw_read and ahw_write */
int delay_statistics[100];

#define MAXENCODE 10

/* structure for mapping of encoding types and sample rates */
/* sample rate == 0 means not supported by af */
int encode_index = 0;
struct enc_t { 
  int audio_enc; /* nevot encoding number */
  char af_enc[10];  /* af encoding  */
  int af_enc_number; /* af encoding number */
  int rate;    /* af supported sample rate */
  int bytes_per_sec; /* bytes per second per channel */
} enc[MAXENCODE] = { {AE_PCMU, "mu-law", 0, 8000, 8000}, 
		     {AE_L16, "pcm-16", 2, 16000, 32000},
                     {AE_L16, "pcm-16", 2, 44100, 88200},
		     {-1, -1, -1, -1, -1}
                   };


/* static functions */

static int ahw_smps2ms(int samples) /* converts sample time to ms */
{ 
  /* sample rate for read and write identical */
  ahw_debug_arg2(9, "ahw: %d samps to %d ms\n", samples, 
       ((1000*samples) / ac->device->playSampleFreq));
  return ((1000*samples) / ac->device->playSampleFreq);
} /* ahw_smps2ms */

static int ahw_ms2smps(int ms) /* converts ms time to sample time */
{
  /* sample rate for read and write identical */
  ahw_debug_arg2(9, "ahw: %d ms to %d samps\n", ms, 
    ((ac->device->playSampleFreq*ms) / 1000)); 
  return((ac->device->playSampleFreq*ms) / 1000);
}

int ahw_open(int ndelay) 
{

  if (first_open) {
    /* default values */
    first_open = 0;
    attributes.play_gain = 10;
    attributes.type = 0;
    attributes.channels = 1;
    attributes.rec_gain = 10;
  }

  /* open connection */
  if (conptr == NULL) {
    if ( (conptr = AFOpenAudioConn(con_nameptr)) == NULL) {
	  fprintf(stderr, "Can't open connection to AF server.\n");
          fprintf(stderr, "Make sure the AF Server is running.\n");
	  exit(1);
    }
    ahw_debug(1, "AF: opened audio connection\n");
  }

  /* open context */
  if (ac == NULL) {
    attributes.endian = ABigEndian;
    ac = AFCreateAC(conptr, af_device, 
       ACEndian | ACPlayGain | ACRecordGain | ACChannels | ACEncodingType, 
               &attributes);
    if (ac->device->playBufType == 2) /* lin16 */{
      ahw_debug(1, "AF: j300 server found");
      device_type = D_J300;
      /* encode_index = 1; */
    }
    else {
      /* encode_index = 0; */
    }
    ahw_debug(1, "AF: created context\n");
  }

   return 0; /* audio_open expects a filedescriptor, we use 0 */
} /* ahw_open */


/*
* Close audio hardware. 
*/
int ahw_close(void)  
{
  if (ac != NULL) {
    AFFreeAC(ac);
    ac = NULL;
    if (ac == NULL) ahw_debug(1, "AF: freed context\n");
  }

  if (ac == NULL) {
    AFCloseAudioConn(conptr);
    conptr = NULL;
    ahw_debug(1, "AF: closed AF connection\n");
  }    
  first_read = 1;
  return 0; /* audio_close expects a filedescriptor, we use 0 */
} /* ahw_close */


/*
* Set audio hardware. Audio device does not have to be open.
* If value < 0, don't change.
* Disable polling signal. Return 0 if ok, -1 if not.
* pr     in   0: record, 1: play
* p      in   parameter to set
* value  in   value to be used; -1: don't change
*/
int ahw_set(int pr, audio_set_t p, int value)
{
  if (conptr == NULL) { 
    ahw_open(0);
  }
  /* switch on different AS types */
  switch(p) {
    case AS_gain:
      /* setting gain is hardcore, because opening and closing server */
      /* is necessary */
      /* opening and closing server is only possible, because */
      /* setting gain is performed by tkaudio, *not* nevot */

      /* set gain to value */
      /* for ingain mapping values 0-255 to 0-24 */
      /* for outgain mapping values 0-255 to -10-18 */
      /*   or -100 to 0, depending on device */
      if(!pr) { /* read */
        ahw_debug_arg1(1, "setting ingain to %d\n", value);
        AFSetInputGain(ac, (int)((float) value / 10.6));
      }
      if(pr == 1) { /* write */
        if (device_type != D_J300) {
          AFSetOutputGain(ac, ((int)((float) value / 9.1) - 10));
        }
        else {
          AFSetOutputGain(ac, ((int)((float)value / 2.55) - 100));
        }
        ahw_debug_arg1(1, "setting outgain to %d\n", value);
      } 
      /* performs the changes ! */
      ahw_close();
      break;
    case AS_port:
      /* ignore, Audiofile always uses all ports available */
      /* could be be better .... */
      break;
    case AS_fillpoint:
      /* ignore, it has no use with AF */
      break;
    case AS_gain_step:
      ahw_debug_arg1(2, "setting gain_step %d\n", value);
      if(!pr) { 
        if((value + ac->attributes.rec_gain <= 24) &&
            (value + ac->attributes.rec_gain >= 0)) {  
          attributes.rec_gain += value;
          AFChangeACAttributes(ac, ACRecordGain, &attributes);
        } 
      }
      if(pr == 1) { 
        if(device_type != D_J300) {
          if((value + attributes.play_gain <= 18) &&
              (value + attributes.play_gain >= -10)) { 
            attributes.play_gain += value;
          }
        }
        else {
          if((value + attributes.play_gain <= 0) && 
              (value + attributes.play_gain >= -100)) { 
            attributes.play_gain += value;
          }
        }
        AFChangeACAttributes(ac, ACPlayGain, &attributes);
        
      }     
      break; 
  }
  return 0;
} /* ahw_set */


/*
* Get current hardware audio status. If an audio error occurred, return
* as quickly as possible to avoid starving the CPU.
* Return -1 if device error, 1 if a play error occurred, 2 if a 
*   record error occurred.
*/
int ahw_get(audio_state_t *as) /* done */
/* not used in main program */  
{
  return 0;
} /* ahw_get */


/*
* Return > 0 if another process is waiting for audio input or output.
* Return -1 if error.
*/
int ahw_waiting(void) /* done */
/* not necessary with audiofile */
{
  return 0;
} /* ahw_waiting */


/*
* Write audio buffer to audio device.
*/
int ahw_write(char *buffer, int bytes)
{
  static counter; /* counter for offset update */
  int i, j, time_delta;
  static int play_offset; /* statistical offset in ms */

  ahw_debug(3, "ahw_write\n");
  ahw_debug_arg1(3, "writebytes %d\n", bytes);
  if (first_write) {
    first_write = 0;
    counter = 1;
    play_offset = ahw_ms2smps(10);
    for(i = 0; i < 100; i++) {
      delay_statistics[i] = 0;
    }
  }
  time_delta = ahw_smps2ms(delta);
  if ((0 <= time_delta) && (time_delta < 100)) {
    delay_statistics[time_delta]++;
  }
  counter++;
  if (counter == 100) { /* calculate new offset */
    i = 99;
    j = 0;
    while ((j < 5) && (i > 1)) {
      j += delay_statistics[i];
      i--;
    }
    play_offset = ++i;
    ahw_debug_arg1(2, "play_offset [ms] %d\n", play_offset);
    for(i= 0; i < 100; i++) {
      delay_statistics[i] = 0;
    }
    counter = 1;
  }
  ahw_debug_arg1(2,"TS(future): %d\n", af_read_end+ahw_ms2smps(play_offset+5)
                  - AFGetTime(ac)); 
  AFPlaySamples(ac, af_read_end+ahw_ms2smps(play_offset + 5), 
         bytes, (unsigned char *)buffer);
  ahw_debug_arg2(3, "Time/read_end %d/%d \n", AFGetTime(ac), af_read_end);
  return bytes;
} /* ahw_write */


/*
* Read audio samples from audio device.
* buffer  out   buffer to receive audio data
* bytes   in    number of audio bytes to read
*/
int ahw_read(char *buffer, int bytes)
{
  ATime t1;
  int af_read_delta;
  ahw_debug(3, "ahw_read\n");
  ahw_debug_arg1(3, "readbytes %d\n", bytes);
  af_read_delta = (bytes*enc[encode_index].rate)/  /* read time [smps] */
                    (enc[encode_index].bytes_per_sec*attributes.channels);
  ahw_debug_arg1(3, "af_read_delta %d\n", af_read_delta);
  if (first_read) { /* first packet after open not read */
    first_read = 0;
    af_read_end = AFGetTime(ac);
    delta = 0;
  }
  else {
    af_read_start = af_read_end;
    af_read_end = af_read_start + af_read_delta;
    if ((af_read_start > AFMAXTIME) || (af_read_end > AFMAXTIME)) {
      af_read_start = 0;
      af_read_end = af_read_delta;
    }
    t1 = AFRecordSamples(ac, 
	       af_read_start, bytes, (unsigned char *)buffer, ABlock);
    ahw_debug_arg1(3, "time-readend %d\n", t1 - af_read_end);
    /* delta between af_read and return of AFRecS. */    
    delta = (t1 > AFMAXTIME) ? 0 : t1 - af_read_end; 
    /* if x-server is blocked, (e.g. resizing windows) */
    /* we have to re-synchronize */
    if (ahw_smps2ms(delta) >= 300) { /* if more than .3 seconds late */
      af_read_end = t1;
    }
  } 
  return bytes;

} /* ahw_read */


/*
 * Returns difference between syncpoint and return value of
 * AFRecordSamples, in bytes.
 */
int ahw_read_queue(int *overflow) 
{
  return delta; /* convert from samples to bytes */
} /* ahw_read_queue */


/* not used with audiofile */
void ahw_track(signal_func_t func)
{

}

int ahw_fmt(audio_descr_t *record, audio_descr_t *play)
{

  ahw_debug_arg3(1, "fmt parms(in): enc/rate/channels: %d/%d/%d\n", 
  record->encoding, record->sample_rate, record->channels);

  /* device type is determined if audio is open */
  if (ac == NULL) {
    ahw_open(0);
  }
  
  /* set audiohardware to best approximation of encoding */
  /* depending on server and it's samplefreq. */

  switch (ac->device->recSampleFreq) {
    
    case 8000:
      record->encoding = AE_PCMU;
      attributes.type = 0; /* mu-law */
      encode_index = 0;
      if (device_type != D_J300) {
        record->channels = 1;
        attributes.channels = 1;
      }
      else {
        attributes.channels = record->channels;
      }
      break;

    case 16000:
      record->encoding = AE_L16;
      attributes.type = 2; /* lin 16 */
      encode_index = 1;
      attributes.channels = record->channels;
      break;

    case 44100:
      record->encoding = AE_L16;
      attributes.type = 2; /* lin 16 */
      encode_index = 2;      
      /* two channels doesn't work */
      attributes.channels = record->channels = 1; 
      break;
  }
  record->sample_rate = ac->device->recSampleFreq;
  if ((device_type == D_J300) && (record->channels == 1)) {
    af_device = 1;
  }
  else {
    af_device = 0;
  }

  ahw_debug_arg1(1, "encode_index %d\n", encode_index);

  ahw_debug_arg3(1, "fmt parms (out): enc/rate/channels %d/%d/%d\n",
       record->encoding, record->sample_rate, record->channels);

  /* changing  attributes */
  /* changing doesn't work we need new contexts */
  AFFreeAC(ac);
    ac = AFCreateAC(conptr, af_device, 
      ACEndian | ACRecordGain | ACPlayGain | ACChannels | ACEncodingType, 
        &attributes);
    ahw_debug_arg1(1, "af_device %d\n", af_device);

  *play = *record;
  return 0;
} /* ahw_fmt */


/* audiofile buffers data, no underflow of buffer */
int ahw_write_queue(int *underflow)
{
  *underflow = 0;
  return 0;
} /* ahw_write_queue */
