/*
gsm -- uniform interface to GSM codec functions.

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include "types.h"      /* u_int8 */
#include "audio_descr.h"
#include "gsm_nevot.h"
#include "pcmu_l.h"
#define GSM_FRAME 160   /* samples per frame */
#define GSM_CODED 33    /* bytes per coded frame */



/*
* Allocate state storage and initialize it.
* Return 0 if no state storage could be allocated.
* gspt     I/O pointer to state to be initialized; 0 if create
* period   I   frame period in msec (ignored)
*/
void *GSM_init(void *state, double period)
{
  struct gsm_state *gspt = (struct gsm_state *)state;
  int one = 1;
  if (!gspt) gspt = gsm_create();
  if (gspt) gsm_option((gsm)gspt, GSM_OPT_FAST, &one);
  return (void *)gspt;
} /* GSM_init */


/*
* Compress audio data in in_buf.
* XXX Doesn't handle multiple channels.
* in_buf    I   input samples
* in_size   I   number of input bytes (not samples)
* ad        I   descriptor of input format
* out_buf   O   buffer to hold encoded audio frame
* out_size  O   size of output buffer in bytes
* state     I/O encoder state
*/
int GSM_encode(void *in_buf, int in_size, audio_descr_t *ad, 
  void *out_buf, int *out_size, void *state, int header_flag)
{
  int i;
  u_int8 *c;            /* input buffer, as u_int8 */
  gsm_signal *s;        /* input buffer, as int16 */
  gsm_signal buf[160];  /* temporary buffer for one frame */

  *out_size = 0;
  if (ad->encoding == AE_PCMU) {
    c = (u_char *)in_buf;
    while (in_size >= 160) {
      for (i = 0; i < 160; i++) buf[i] = pcmu_l16(*c++);
      gsm_encode((gsm)state, (gsm_signal *)buf, (gsm_byte *)out_buf);
      in_size -= 160; 
      *out_size += GSM_CODED;
    }
  }
  else {
    s = (gsm_signal *)in_buf;
    while (in_size >= 320) {
      gsm_encode((gsm)state, s, (gsm_byte *)out_buf);
      in_size -= 320;
      s += 160;
      *out_size += GSM_CODED;
    }
  }
  return in_size == 0 ? 0 : -1; 
} /* GSM_encode */


/*
* Output buffer must be large enough to hold result as words.
* in_buf     I   input samples (encoded)
* in_size    I   number of input bytes
* ad         I   descriptor of input format
* out_buf    O   buffer to hold encoded audio frame
* out_size   O   size of output buffer in bytes
* state      I/O encoder state
*/
int GSM_decode(void *in_buf, int in_size, audio_descr_t *ad, 
  void *out_buf, int *out_size, void *state)
{
  int i, k;
  int frames;       /* number of input frames */
  u_int8 *oc, *ic;  /* buffer pointer for byte I/O buffers */
  gsm_signal *s;    /* output */
  int status = 0;

  frames = in_size / GSM_CODED;
  ic = (u_int8 *)in_buf;
  oc = (u_int8 *)out_buf;
  s  = (gsm_signal *)out_buf;
  for (k = 0, *out_size = 0; k < frames;
       k++, ic += GSM_CODED, s += GSM_FRAME) {
    status = gsm_decode((gsm)state, (gsm_byte *)ic, s);
    if (status < 0) return status;

    /* convert 16-bit linear to mu-law (in place) if necessary */
    if (ad->encoding == AE_PCMU) {
      for (i = 0; i < GSM_FRAME; i++) *oc++ = l16_to_pcmu(s[i]);
      *out_size += GSM_FRAME;
    }
    else {
      *out_size += sizeof(gsm_signal)*GSM_FRAME;
    }
  }
  return status; 
} /* GSM_decode */
