/*
linux_audio  --  audio interface for Linux (586)

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <fcntl.h>       /* O_RDWR */
#include <sys/ioctl.h>   /* ioctl() */
#include <unistd.h>      /* close(), ioctl() */
#include <signal.h>      /* signal() */
#include "audio_generic.h"
#include "ahw.h"
#include <errno.h>
#include <ansi.h>

#define AUDIO_DATA "/dev/audio"
#define AUDIO_CTL  "/dev/audioctl"



static int ac = -1;                 /* audio control port */
static int au = -1;                 /* audio data port */
typedef void (*signal_func_t)(int);
static signal_func_t track;         /* handler that tracks audio state */
static long play_count, rec_count;  /* number of bytes played, recorded */


/*
* Signal handler.  Call actual handler and then reestablish signal
* handler.
*/
static void ahw_signal(int sig)
{
  (*track)(sig);
  signal(SIGPOLL, ahw_signal);
} /* ahw_signal */

/*
* Open audio control device. The BSD audio driver doesn't have one, so
* ac and au_* are treated as one and the same in that case.
* Also, set up SIGPOLL handler if track_audio_state flag is TRUE.
* Return -1 on error, 0 on success.
*/
static int ahw_open_control(void)
{
  if (ac < 0) {
  }
  return 0;
} /* ahw_open_control */


/*
* Open audio hardware in given mode 'iomode' and set file descriptor
* to non-blocking.  Return file descriptor number or -1 on error.  If
* the output device is busy, we don't fail, but would try again later,
* thus grabbing the audio device as quickly as possible.  This allows
* short announcements (e.g., biff) to grab the audio device without
* crashing this program.  If the other program changes the device
* characteristics, there is potential for disaster.  Several nevot's
* can run concurrently if the nevot.talk is used to disable the
* recording device (iff there is no audio output waiting).
*/
int ahw_open(int iomode)
{
  /* open audio data stream */
  if (!iomode) return -1;

  return au;
} /* ahw_open */


/*
* Close audio hardware. If (iomode & A_WR), close writing.
* If (iomode & A_RD), close reading. Return old (now closed) input
* file descriptor to allow disabling input handlers.
*/
int ahw_close(int iomode)
{
  int au_old = au;

  if (au >= 0) {
    close(au);
    au = -1;
  }
  return au_old;
} /* ahw_close */


/*
* Set audio hardware. Audio device does not have to be open.
* If value < 0, don't change.
* Disable polling signal. Return 0 if ok, -1 if not.
* pr     in   0: record, 1: play, 2: monitor
* p      in   parameter to set
* value  in   value to be used; -1: don't change
*/
int ahw_set(int pr, audio_set_t p, int value)
{
  int success = -1;    /* return status */
  int adev;            /* audio device to use */

  /* nothing to change */
  if (value < 0) return 0;

  /* determine whether to set through control device or actual audio */
  if (p == AS_port || p == AS_gain) {
    if (ahw_open_control() < 0) return -1;
    adev = ac;
  }
  else adev = au;

  switch(p) {
  case AS_port:
    if (pr == 0) {
      if      (value == AP_in_line)       return -1;
      else if (value == AP_in_microphone) return -1;
      else return -1;
    }
    else {
      if (value & AP_out_speaker)   return -1;
      if (value & AP_out_headphone) return -1;
      if (value & AP_out_line)      return -1;
      if (value & ~(AP_out_speaker|AP_out_headphone|AP_out_line))
        return -1;
    }
    break;

  case AS_gain:
    break;

  case AS_gain_step:
    break;

  case AS_rate:
    break;

  case AS_channels:
    break;

  case AS_encoding:
    break;

  case AS_fillpoint:
    break;

  /* audio state tracking */
  case AS_track:
    if (track == (signal_func_t)value) break;
    track = (signal_func_t)value;
    if (track) {
      signal(SIGPOLL, ahw_signal);
    }
    else {
      signal(SIGPOLL, SIG_IGN);
    }
    return 0;  /* nothing to set via ioctl() */
  } /* switch */

  /* actually set parameter if valid audio device */
  if (adev > 0) {
  }
  return success;
} /* ahw_set */

/*
* Return number of bytes in write queue.
* If the number is negative, an underflow has occurred.
* Return -1 if error.
*/
int ahw_write_queue(int *errors)
{
  return 0;
} /* ahw_write_queue */


/*
* Return occupancy of read queue in bytes.
*/
int ahw_read_queue(int *errors)
{
  return 0;
} /* ahw_read_queue */

/*
* Get current hardware audio status.
* Return -1 if error, 0 if success.
*/
int ahw_get(audio_state_t *as)
{
  if (ahw_open_control() < 0) return -1;

  return 0;
} /* ahw_get */

/*
* Return > 0 if another process is waiting for audio input or output.
* Return -1 if error.
*/
int ahw_waiting(void)
{
  return 0;
} /* ahw_waiting */

/*
* Write audio buffer to audio device.
*/
int ahw_write(char *buffer, int bytes)
{
  if (bytes == 0) return bytes;
  if (au >= 0) {
    bytes = write(au, buffer, bytes);
    if (bytes > 0) play_count += bytes;  
  }
  return bytes;
} /* ahw_write */


/*
* Read audio samples from audio device.
* buffer  out   buffer to receive audio data
* bytes   in    number of audio bytes to read
*/
int ahw_read(char *buffer, int bytes)
{
  bytes = read(au, buffer, bytes);
  if (bytes > 0) rec_count += bytes;  
  return bytes;
} /* ahw_read */
