/*
mbuf  --  mbuf-like buffer structures.

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include <sys/types.h>
#include "mbuf.h"
#include <stdio.h>     /* printf() */
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <ansi.h>



static mbuf_t *pool;   /* beginning of buffer pool chain */
static int available;  /* number of available mbuf's */
static int mbuf_size;  /* size of total mbuf */

/*
* Check whether the claimed number of mbufs is in pool.
*/
void m_check(mbuf_t *not)
{
  int count;
  mbuf_t *m;

  printf("m_chk");
  for (count = 0, m = pool; m; m = m->m_next, count++) {
    printf(" %p", m);
    if (m == not) {
      printf("mbuf allocated %p in free list\n", m);
      goto error;
    }
    if (m->m_type != MT_FREE) {
      printf("mbuf %p type %d\n", m, m->m_type);
      goto error;
    }
    if ((m == m->m_next) || (count > 1000)) {
      printf("mbuf loop %p (%d)\n", m, count);
      goto error;
    }
  }
  if (available != count) {
    printf("mbuf available %d count %d\n", available, count);
    for (m = pool; m; m = m->m_next) {
      printf("%p ", m);
    }
    printf("\n");
    goto error;
  }
  printf("\n");
  return;

error:
  return;
} /* m_check */

/*
* Obtain a buffer from the pool; if none available, fail and return 0.
* Allocated space is 32-bit aligned. Initial length is zero.
*/
mbuf_t *m_get(char *who)
{
  mbuf_t *m;

  m    = pool;
  pool = pool->m_next;
  m->m_off  = MMINOFF(m);
  m->m_len  = 0;
  m->m_next = NULL;
  m->m_type = 0;
  m->m_ref  = 0;
  available--;
#if DEBUG
  printf("m_get %p who=%s pool=%p avail=%d\n", m, who, pool, available);
#endif
 
  return m;
} /* m_get */

/*
* Decrement reference count; if <= 0, return buffer to free pool.
*/
void m_free(mbuf_t *m, char *who)
{
  m->m_ref--;
  if (m->m_ref <= 0) {
    m->m_next = pool;
    assert(m->m_type != MT_FREE);
    m->m_type = MT_FREE;
    pool = m;
    available++;
#if DEBUG
    printf("m_fre %p who=%s pool=%p avail=%d\n", m, who, pool, available);
#endif
  }
} /* m_free */


/*
* Return chain of buffers to free pool.
*/
void m_freem(mbuf_t *mh, char *who)
{
  mbuf_t *m, *m_next;

  for (m = mh; m; m = m_next) {
    m_next = m->m_next; 
    m_free(m, who);
  }
} /* m_freem */


/*
* Create a buffer pool of 'count' mbufs, each of size 'alloc'. 
* Storage is allocated consecutively to limit malloc header overhad.
*/
int m_create(int count, size_t alloc)
{
  int i;
  mbuf_t *m, *m_next = 0;

  mbuf_size = (alloc + sizeof(mbuf_t)) & (~07);
  if (pool) return mbuf_size;
  pool = (mbuf_t *)calloc(count, mbuf_size);
  assert(pool);

  for (i = 0, m = pool; i < count; i++, m = m_next) {
    m_next      = (mbuf_t *)((char *)m + mbuf_size);
    m->m_next   = (i == count-1) ? 0 : m_next;
    m->m_type   = MT_FREE;
  }
  available = count;
  return alloc;
} /* m_create */


/*
* Create exact replica of input mbuf 'm', occupying separate storage.
*/
mbuf_t *m_copy(mbuf_t *m)
{
  mbuf_t *c; /* new mbuf */

  c = m_get("mc");
  *c = *m; /* copy header, including m_off */
  memcpy(mtod(c, char *), mtod(m, char *), m->m_len);
  return c;
} /* m_copy */
