/*
tsap  --  network address objects

Copyright 1993 by AT&T Bell Laboratories; all rights reserved
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>        /* AF_UNIX domain */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rpcsvc/ypclnt.h>
#include <memory.h>        /* memcpy.h */
#include <netdb.h>         /* h_name */
#include <stdlib.h>
#include <sys/param.h>     /*& MAXHOSTNAMELEN */
#include <stdio.h>         /* sprintf() */
#include <string.h>
#include <sys/utsname.h>   /* uname() */

#include "tsap.h"
#include <ansi.h>



#undef sun

/*
* Allocate the memory area within a tsap pointed to by 't'. Does NOT
* allocate the tsap_t structure itself within this routine.
*/
void tsap_alloc(tsap_t *t, int family)
{
  t->family = family;

  switch (family) {
  case T_sockaddr:
    t->addr.buf    = malloc(sizeof(struct sockaddr));    
    t->addr.maxlen = sizeof(struct sockaddr);
    t->addr.len    = t->addr.maxlen;
    break;

  case T_unix:
    t->addr.buf    = malloc(sizeof(struct sockaddr_un));    
    t->addr.maxlen = sizeof(struct sockaddr_un);
    t->addr.len    = t->addr.maxlen;
    break;

  case T_osi:
    break;

  /* maximum size if currently unknown */
  case T_unknown:
    t->addr.buf    = malloc(sizeof(struct sockaddr_un));    
    t->addr.maxlen = sizeof(struct sockaddr_un);
    t->addr.len    = t->addr.maxlen;
    break;
  } 
} /* tsap_alloc */


/*
* Frees data area within a tsap datastructure; does NOT free the memory
* region pointed to by 't' itself.
*/
void tsap_free(tsap_t *t)
{
  if (t->addr.buf) free(t->addr.buf);
  t->addr.maxlen = 0;
  t->addr.len    = 0;
  t->family      = 0;
} /* tsap_free */


/* 
* Copy with new allocation.
*/
void tsap_cpy(tsap_t *d, tsap_t *s)
{
  if (d->addr.maxlen < s->addr.len) {
    if (d->addr.len) tsap_free(d);
    tsap_alloc(d, s->family);
  }
  else d->family = s->family;
  d->addr.len = s->addr.len;
  memcpy(d->addr.buf, s->addr.buf, s->addr.len);
} /* tsap_cpy */


/*
* Compare address and port number. Return 1 if equal.
*/
int tsap_equ(tsap_t *a, tsap_t *b)
{
  struct sockaddr_in *sin_a, *sin_b;
  struct sockaddr_un *sun_a, *sun_b;

  if (a->family != a->family) return 0;
  switch(a->family) {
    case T_sockaddr:
      sin_a = TSAP_SINP(a);
      sin_b = TSAP_SINP(b);
      switch(sin_a->sin_family) {
        case AF_INET:
          return (sin_a->sin_addr.s_addr == sin_b->sin_addr.s_addr &&
                  sin_a->sin_port        == sin_b->sin_port);
      }
      break;

    case T_unix:
      sun_a = TSAP_SUNP(a);
      sun_b = TSAP_SUNP(b);
      return strcmp(sun_a->sun_path, sun_b->sun_path) == 0;
      break;

    case T_osi:
      break;

    case T_unknown:
      return a->addr.len == b->addr.len &&
             memcmp(a->addr.buf, b->addr.buf, a->addr.len) == 0;
  }
  return 0;
} /* tsap_equ */


/*
* Compare address only.
*/
int tsap_equ_a(tsap_t *a, tsap_t *b)
{
  struct sockaddr_in *sin_a, *sin_b;

  if (a->family != a->family) return 0;
  switch(a->family) {
    case T_sockaddr:
      sin_a = TSAP_SINP(a);
      sin_b = TSAP_SINP(b);
      switch(sin_a->sin_family) {
        case AF_INET:
          return sin_a->sin_addr.s_addr == sin_b->sin_addr.s_addr;
      }

    case T_unknown:
    case T_unix:
      return tsap_equ(a, b);
      break;

    case T_osi:
      break;
  }
  return 0;
} /* tsap_equ_a */


/*
* Convert tsap to ASCII. Return pointer to static area.
*/
char *tsap_ntoa(tsap_t *t)
{
  struct sockaddr_in *sin;
  struct sockaddr_un *sun;

  switch(t->family) {
    case T_sockaddr:
      sin = TSAP_SINP(t);
      switch(sin->sin_family) {
        case AF_INET:
          return inet_ntoa(sin->sin_addr);
      }

    case T_unix:
      sun = TSAP_SUNP(t);
      return sun->sun_path;
      break;

    case T_osi:
      return "";
      break;

    case T_unknown:
      return "";
      break;
  }
  return 0;
} /* tsap_ntoa */


/*
 * Return 1 if t is a broadcast address.
 */
int tsap_isbroadcast(tsap_t *t)
{
  u_int32 ipaddr;
  u_int32 class, host, mask;
  struct sockaddr_in *sin;
   
  switch(t->family) {
    case T_sockaddr:
      sin = TSAP_SINP(t);
      ipaddr = sin->sin_addr.s_addr;
      class  = (ipaddr >> 28) & 0xd;
      mask   = (class == 0)   ? (0xffffff):     /* class a */
               (class == 0x8) ? (0xffff):       /* class b */
               (0xff);                          /* class c */
      /* *** what about subnet mask? *** */
      host = (ipaddr & mask);
      return (host == 0) || (host == mask); 

    case T_unknown:
    case T_unix:
      return 0;

    case T_osi:
      return 0;
  }
  return 0;
} /* tsap_isbroadcast */


/*
* Return 1 if 't' is a multicast address.
*/
int tsap_ismulticast(tsap_t *t)
{
  u_int32 ipaddr;
  struct sockaddr_in *sin;
   
  switch(t->family) {
    case T_sockaddr:
      sin = TSAP_SINP(t);
      ipaddr = sin->sin_addr.s_addr;
      return IN_MULTICAST(ntohl(ipaddr));

    case T_unknown:
    case T_unix:
      return 0;
 
    case T_osi:
      return 0;
  }
  return 0;
} /* tsap_ismulticast */


/*
* Get address of own host, of family 'family' and write it to
* 'tsap'. The necessary buffer for that address is allocated
* by the routine and must be freed by tsap_free() when no longer
* needed.
*/
void tsap_gethostaddress(int family, tsap_t *tsap)
{
  struct hostent *hep;
  struct utsname name;

  uname(&name);
  if (family == AF_INET) {
    struct sockaddr_in *sin;

    tsap_alloc(tsap, T_sockaddr);
    hep = gethostbyname(name.nodename);
    if (!hep) return;
    if (hep->h_addrtype == AF_INET) {
      sin             = TSAP_SINP(tsap);
      sin->sin_family = AF_INET;      
      sin->sin_addr.s_addr = *(u_int32 *)(*hep->h_addr_list);
    }
  }
} /* tsap_gethostaddress */


/*
* Copy network address 't' into buffer 'buf'. Set length field 'len'.
* Return address family.
*/
int tsap_pullup(char *buf, tsap_t *t, int *len)
{
  struct sockaddr_in *sin;
  struct sockaddr_un *sun;

  switch(t->family) {
    case T_sockaddr:
      if (buf) {
        sin = TSAP_SINP(t);
        memcpy(buf, (char *)&sin->sin_addr.s_addr, sizeof(long));
      }
      *len = 4;
      return 1;

    case T_unix:
      sun = TSAP_SUNP(t);
      strcpy(buf, sun->sun_path);
      return strlen(sun->sun_path);

    case T_unknown:
      memcpy(buf, t->addr.buf, t->addr.len);
      return t->addr.len;

    case T_osi:
      return 22;
  }
  return 0;
} /* tsap_pullup */


/*
* Return tsap port number.
*/
int tsap_getport(tsap_t *t)
{
  struct sockaddr_in *sin;

  switch(t->family) {
    case T_sockaddr:
      sin = TSAP_SINP(t);
      return sin->sin_port;

    case T_unknown:
    case T_unix:
      return 0;

    case T_osi:
      return 0;

    default:
      return 0;
  }
} /* tsap_getport */


/*
* Set port number of tsap 't' to 'port' (host byte order).
*/
void tsap_setport(tsap_t *t, u_int32 port)
{
  struct sockaddr_in *sin;

  switch(t->family) {
    case T_sockaddr:
      sin = TSAP_SINP(t);
      sin->sin_port = htons(port);
      break;

    case T_unknown:
    case T_unix:
      break;

    case T_osi:
      break;
  }
} /* tsap_setport */


/*
* Translate host name 'name' into network addresses (typically one,
* but may be more if single host has several aliases, e.g., several
* interfaces or is reachable through several network layer protocols). 
* The port number of the TSAPs is left unmodified.  Return number of
* valid addresses, -1 on failure.
*/
int tsap_gethostbyname(char *name, tsap_t adrlist[], int listlen)
{
  struct sockaddr_in *sin;
  register struct hostent *hep;
  u_int32 ipaddr;   
  register int i, count = 1;
  char **ap;

  if (!name || strlen(name) == 0) {
    count = 0;
  }
  /* attempt dotted decimal */
  else if ((ipaddr = inet_addr(name)) != -1) {
    adrlist[0].family    = T_sockaddr;
    tsap_alloc(&adrlist[0], T_sockaddr);
    sin                  = TSAP_SIN(adrlist[0]);
    sin->sin_family      = AF_INET;
    sin->sin_addr.s_addr = ipaddr;
  }
  /* attempt host name */
  else if ((hep = gethostbyname(name))) {
    ap = hep->h_addr_list;
    for (i = count = 0; i < listlen && *ap; i++) {
      count++;
      tsap_alloc(&adrlist[i], T_sockaddr);
      sin                  = TSAP_SIN(adrlist[i]);
      sin->sin_family      = AF_INET;
      sin->sin_addr.s_addr = *(u_int32 *) *ap++;
    }
  }
  /* last resort: YP lookup */
  else {
    static char *domain = 0;  /* YP domain */
    char *value;              /* key value */
    int value_len;            /* length of returned value */

    if (!domain) yp_get_default_domain(&domain);
    count = yp_match(domain, "hosts.byname", name, strlen(name),
      &value, &value_len);
    if (count == 0) {  /* YP lookup was successful */
      tsap_alloc(&adrlist[0], T_sockaddr);
      sin                  = TSAP_SIN(adrlist[0]);
      sin->sin_family      = AF_INET;
      sin->sin_addr.s_addr = inet_addr(value);
      free(value);
      count = 1;
    }
    else count = -1;
  }

  return count;
} /* tsap_gethostbyname */


/*
* Given network address, return pointer to statically allocated host
* name string (domain name format, e.g., foo.bar.com). Return 0 if
* no translation.
*/
char *tsap_gethostbyaddr(tsap_t *t)
{
  struct hostent *hep;
  struct sockaddr_in *sin;

  if (!t) return NULL;
  sin = TSAP_SINP(t);

  if (t->family == T_sockaddr && sin->sin_family == AF_INET) {
    hep = gethostbyaddr((char *)&sin->sin_addr.s_addr,
       sizeof(u_int32), AF_INET);
    if (!hep) return inet_ntoa(sin->sin_addr);
    else return hep->h_name;
  }
  return (char *)0;
} /* tsap_gethostbyaddr */


/*
* Print tsap to internal buffer. %p: port; %h: host name; %n: host number.
*/
char *tsap_sprintf(char *buf, char *fmt, tsap_t *t)
{
  static char buffer[256], *b;
  char *f;
  char *h;   /* host name */
  int flag = 0;

  if (!buf) buf = buffer;
  b = buf;

  for (f = fmt; *f; f++) {
    if (flag) {
      flag = 0;
      switch(*f) {
        case 'p':
          if (t->family == AF_INET) sprintf(b, "%d", tsap_getport(t));
          break;

        case 'h':
          h = tsap_gethostbyaddr(t);
          if (h) {
            strcpy(b, h);
            break;
          }
          /* fall through if no name */

        case 'n':
          strcpy(b, tsap_ntoa(t));
          break;

        default:
          *b++ = *f;
          break;
      }
      if (strchr("hnp", *f)) b += strlen(b);
    }
    else {
      if (*f == '%') flag = 1;
      else *b++ = *f;
    }
  }
  *b = '\0';
  return buf;
} /* tsap_sprintf */


/*
* Return IPv4 address as 32-bit integer.
*/
u_int32 tsap_ip4(tsap_t *tsap)
{
  return TSAP_SINP(tsap)->sin_addr.s_addr;
} /* tsap_ip4 */
