/*
idstring  --  generate local identification, sprintf-like.

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <ansi.h>        /* architecture-specific definitions */ 
#include <stdio.h>       /* L_cuserid */
#include <stdlib.h>      /* malloc() */
#include <string.h>      /* strcpy() */
#include <ctype.h>       /* isspace() */
#include <pwd.h>         /* struct passwd */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>  /* struct inaddr */
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/utsname.h> /* uname() */
#include <unistd.h>      /* getuid(), ttyname() */
#include <time.h>        /* localtime() */
#include "idstring.h"    /* getinetdomain() */



/*
* Extract part of name, with parts separated by dots:
* if part > 0, use last 'part' pieces; if part < 0, use all but 'part'
* pieces.
*/
static char *name_part(char *name, int part)
{
  char *p[10];
  char *s;
  int count = 0;

  if (part == 0) return name;

  for (s = name; *s; s++) {
    if (*s == '.') p[count++] = s+1;
  }
  if (part > 0) {
    return (count - part) >= 0 ? p[count - part] : name; 
  }
  else {
    return (-part-1) >= 0 ? p[-part-1] : ""; 
  }
} /* name_part */


/*
* item:   item within GECOS field
* dest:   destination buffer
* length: space length
*/
static void cusername(int item, char *dest, int length)
{
  struct passwd *pw;
  char *s;   /* source */
  char *d;   /* destination */
  int i;

  pw = getpwuid(getuid());
  if (pw) {
    /* select item */
    s = pw->pw_gecos;
    for (i = 0; i < item && s; i++) {
      s = strchr(s, ',');
      if (s) s++;  /* skip comma */
    }

    for (d = dest; s && *s && (length > 0) && (*s != ',' && *s != ':'); s++) {
      if (*s == '&' && length > L_cuserid) {
        cuserid(d);
        length -= strlen(d);
        d += strlen(d);
      }
      else {
        *d++ = *s;
        length--;
      }
    }
    *d = '\0';
  }
} /* cusername */


/*
* Convert 'John G. Doe' into 'John.G.Doe' as in mailing
* addresses.
*/
static void dotted(char *b)
{
  char *s, *d;
  int skip = 0, first_space = 0;

  for (s = d = b; *s; s++) {
    if (*s == '.') {
      skip = 1;
      *d++ = *s;
    }
    else if (isspace(*s)) {
      if (!skip && first_space) *d++ = '.';
      first_space = 0;
    }
    else {
      skip = 0;
      first_space = 1;
      *d++ = *s;
    }
  }
  *d = '\0';
} /* dotted */


/*
* Return pointer to dynamically allocated string of size 'alloc'
* according to format 'fmt'.  The following special characters are
* recognized and replaced as shown below:
* %%  the percent sign itself
* %h  host name (foo)
* %n  real name of user (John Doe)
* %N  real name, with dots (John.Doe)
* %o  office
* %d  domain name (foo.ee.bar.com)
* %D  domain name; like %d, but w/O host (ee.foo.bar)
* %nD domain name; only n rightmost parts (n=2: foo.bar)
* %e  office phone extension
* %p  home phone
* %u  user login name
* %i  numeric network address (e.g., Internet number)
* %t  tty name
* %Z  time zone name
* %z  offset from GMT in hours
* %$[A-Z]* environment variable of given name (until space)
*/
char *idstring(char *fmt, int alloc)
{
  char *b;                 /* buffer */
  int space_left;          /* bytes left within allocated area */
  char *f;                 /* pointer within format */
  char *s;                 /* temporary */
  char *d;                 /* destination */
  int arg = 0;             /* field width */
  int flag = 0;            /* within % field */
  char *eos;               /* end of string */
  char save;               /* character at eos (saved) */
  time_t clock;
  struct utsname name;     /* uname value */
  struct hostent *hep;
  struct tm *tm;
  extern int gmtoff(void);

  if (!fmt) return (char *)0;
  b = malloc(space_left = alloc);
  if (!b) return b;
  *b = '\0';

  for (f = fmt, d = b; *f && space_left > 1; f++) {
    if (*f == '%') {
      arg  = 0;
      flag = 1;
    }
    else if (flag) {
      switch(*f) {
      case '-':
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
        arg = strtol(f, &f, 10);
        f--;
        flag++;
        break;

      case 'd':  /* fully qualified domain name */
        getinetdomain(d, space_left);
        break;

      case 'D':  /* without host part, various parts */
        getinetdomain(d, space_left);
        /* overlapping copy! */
        {
          char *s = name_part(d, arg), *p = d;

          while (*s) *p++ = *s++;
          *p = '\0';
        }
        break;

      case 'e':  /* extension */
        cusername(2, d, space_left);
        break;

      case 'h':  /* host name */
        uname(&name);
        if (space_left > strlen(name.nodename)) strcpy(d, name.nodename);
        break;

      case 'i':  /* own Internet number */
        uname(&name);
		if ((hep = gethostbyname(name.nodename)) && hep->h_addrtype == AF_INET) {
          if (space_left > 15) {
            struct in_addr in;
            in.s_addr = *(u_long *)(*hep->h_addr_list);
            strcpy(d, inet_ntoa(in));
          }
        }
        break;

      case 'n':  /* user real name */
      case 'N':
        cusername(0, d, space_left);
        if (*f == 'N') dotted(d);
        break;

      case 'o':  /* office */
        cusername(1, d, space_left);
        break;

      case 'p':  /* home phone */
        cusername(3, d, space_left);
        break;

      case 't':  /* tty name */
        s = ttyname(0);
        if (s && strlen(s) < space_left) strcpy(d, s);
        break;
      
      case 'u':  /* user login name */
        if (space_left > L_cuserid) cuserid(d);
        break;

      case 'Z':  /* time zone name */
        time(&clock);
        tm = localtime(&clock);
        /* tm_zone is not supported in SYSV, HP-UX */
        strftime(d, space_left, "%Z", tm);
        break;

      case 'z': /* GMT offset in hours */
        if (space_left > 6) 
          sprintf(d, "%+.2d00", gmtoff() / 3600);
        break;

      case '$':  /* environment variable */
        /* search for end of environment variable */
        for (eos = f+1; *eos && !isspace(*eos); eos++) {};
        save = *eos;
        if (*eos) *eos = '\0';
        if (getenv(f+1)) strncpy(d, getenv(f+1), space_left);
        *eos = save;
        f = eos-1; 
        break;

      case '\0':
        f--;
        break;
        
      default:
        *d = *f;
        *(d+1) = '\0';
        break;
      }
      flag--;
    }
    else {
      *d = *f;
      *(d+1) = '\0';
    }
    d += strlen(d);
    space_left = alloc - (d - b);
  }
  space_left--;
  b = realloc(b, alloc - space_left);
  return b;
} /* idstring */

#include <tcl.h>
#include <stdlib.h>  /* malloc() */

int Tcl_idstring(ClientData client, Tcl_Interp *interp, int argc, 
  char *argv[])
{
  if (argc != 2) {
    sprintf(interp->result, "%s: illegal # of args (%d)", argv[0], argc);
    return TCL_ERROR;
  }

  Tcl_SetResult(interp, idstring(argv[1], 200), TCL_DYNAMIC);
  return TCL_OK;
} /* Tcl_idstring */
