/*
member  --  member-specific routines

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include "nevot.h"
#include <time.h>            /* time() */
#include <stdlib.h>          /* free(), calloc() */
#include <string.h>          /* strdup() */
#include <sys/time.h>        /* gettimeofday() */
#include "util.h"            /* strsaven() */
#include "audio_generic.h"   /* audio_close() */
#include "debug.h"
#include "member.h"
#include <assert.h>
#include <ansi.h>



static void member_announce(Audio_member_t m);

/*
* Change current member identification (user, host, etc.) if it doesn't
* agree with the one currently displayed. If 'len' is zero, but data
* is non-zero, compute length as strlen(data).
*/
void member_sdes(Audio_member_t m, const SDES_t type, const char *data,
  int len)
{
  int display = FALSE;   /* flag: change member display */
  struct sdes *sdes, *sdes_last = 0;

  if (!m) return;
  /* Search for matching entry or point to next free if none. */
  for (sdes = m->sdes; sdes; sdes = sdes->next) {
    if (sdes->type == type) break;
    sdes_last = sdes;
  }
  /* Create a new SDES entry. */
  if (!sdes) {
    sdes = (struct sdes *)malloc(sizeof(struct sdes));
    if (sdes_last) sdes_last->next = sdes;
    else m->sdes = sdes;
    sdes->next  = 0;
    sdes->value = 0;
    sdes->len   = 0;
    sdes->type  = type;
  }

  if (data && !len) len = strlen(data);

  /* If no current value, set. */
  if (!sdes->value) {
    if (data) {
      sdes->value = strsaven(data, len);
      sdes->len = strlen(sdes->value);
      display  = TRUE;
    }
  }
  /* If current value, but length or content have changed, set to new. */
  else if (data && ((sdes->len != len) || strncmp(sdes->value, data, len))) {
    free(sdes->value);
    sdes->value = strsaven(data, len);
    sdes->len   = len;
    display     = TRUE;
  }
  /* If we got CNAME, we might announce a new member. */
  if (type == SDES_CNAME) member_announce(m);

  /* Show only if our identity is known. */
  if (display && m->id) (*nevot.event)(m->s, m, type, sdes->value);

  if (m->soft_to) {
    m->soft_to = FALSE;
    (*nevot.event)(m->s, m, Audio_alive, 1);
  }
  debug(DEBUG_CNTL, "member_set_sdes %d %s: %s %s", 
    type, tsap_ntoa(&m->nsrc[0]), sdes->value ? sdes->value : "(null)", display ? "D" : "");
} /* member_sdes */

/*
* Announce new member to the outside world, i.e., conference controller.
* Called when CNAME is known. Sets m->id.
*/
static void member_announce(Audio_member_t m)
{
  Audio_session_t s = m->s;
  char *cname;

  cname = Audio_MemberGetSDES(m, SDES_CNAME);
  if ((m->id == 0) && cname) {
    m->id = strdup(cname);
    (*nevot.event)(s, m, Audio_hello);

    /* only now activate listen mode */
    Audio_MemberListen(m, TRUE);
  }
} /* member_announce */

/*
* Create the descriptor for a new member; create network association.
* type      in   control or data session
* nsrc      in   network source address
* src       in   content/synchronization source;
*                if zero, don't add to hash table
*/
static Audio_member_t member_add(Audio_session_t s, int type,
  tsap_t *nsrc, src_id_t src)
{
  Audio_member_t m;
  int new;
  Tcl_HashEntry *entry;
  int i;

  if (s->no_more) return 0;  /* don't allow additional member */
  m = (Audio_member_t)calloc(1, sizeof(struct Audio_member_t));
  m->s = s;
  s->members++;

  /* insert at beginning of double-linked list attached to session */
  m->prev = 0;
  m->next = s->first;
  if (s->first) s->first->prev = m;
  s->first = m;

  /* if src is non-zero, add to hash table */
  if (src) {
    entry = Tcl_CreateHashEntry(&s->member_table, (char *)(long)src, &new);
    /* bug check */
    assert(new);
    Tcl_SetHashValue(entry, m);
  }
  tsap_cpy(&m->nsrc[type & 1], nsrc);

  /* Create valid destination addresses. */
  for (i = 0; i < 2; i++) {
    tsap_cpy(&m->ndst[i], nsrc);
    tsap_setport(&m->ndst[i], s->sp.port + i);
  }

  m->id              = 0;
  m->src             = src;
  m->active          = 0;
  m->listen          = 0;
  m->last_sample     = 0;
  m->sync.delay      = 0;
  m->audio_header    = -1;  /* invalid value */
  m->decoder_state   = 0;
  m->audio           = s->audio;
  m->created         = time(0);
  m->sdes            = 0;

 
  return m;
} /* member_add */

/*
* Remove SDES information.
*/
Audio_errno_t Audio_MemberClose(Audio_member_t m, char *reason)
{
  Audio_session_t s = m->s;
  struct sdes *sdes, *sdes_next;

  /* Remove member from talking indicator list. */
  talking_delete(m);

  /* Let others know, but only if this is an identified source. */
  if (m->id) (*nevot.event)(s, m, Audio_bye, reason);

  /* Free any per-member buffers. */
  for (sdes = m->sdes; sdes; sdes = sdes_next) {
    if (sdes->value) free(sdes->value);
    sdes_next = sdes->next;
    free(sdes);
  }
  m->sdes = 0;
  if (m->id) {
    free(m->id);
    m->id = 0;
  }
  return Audio_OK;
} /* Audio_MemberClose */

/*
* Destroy member pointed to be 'm', unlinking it from session chain.
* We must clear the session socket cache as well.
* Return Audio_OK on success, -1 on error.
*/
Audio_errno_t Audio_MemberDelete(Audio_member_t m, char *reason)
{
  Audio_session_t s, ss; /* session pointer */
  boolean alive;
  Tcl_HashEntry *entry;
  int i;

  if (!m) return 0;
  s = m->s;

  Audio_MemberClose(m, reason);

  s->members--;
  if (m->decoder_state) free(m->decoder_state);

  /* unlink member descriptor from doubly linked list */ 
  if (m->prev) m->prev->next = m->next;
  if (m->next) m->next->prev = m->prev;
  /* if first in list, correct session list */
  if (m == s->first) s->first = m->next;

  /* delete entry in hash table if src is non-zero */
  if (m->src) {
    entry = Tcl_FindHashEntry(&s->member_table, (char *)(long)m->src);
    Tcl_DeleteHashEntry(entry);
  }

  /* make sure cache doesn't point to this member */
  if (s->cache == m) s->cache = 0;

  /* free member-specific data structures */
  for (i = 0; i < 2; i++) {
    tsap_free(&m->nsrc[i]);
    tsap_free(&m->ndst[i]);
  }
  free(m);

  /* check whether there are any sessions with members left */
  alive = FALSE;
  for (ss = nevot.first; ss && !alive; ss = ss->next) {
    alive = (ss->first != 0);
  }
  if (!alive) {
    nevot.audio.talking = FALSE;   /* no need to update display - none left! */
    audio_close();
  }
  return Audio_OK;
} /* Audio_MemberDelete */

/*
* Return SDES value for member 's' in session 's'.  Return "" if not
* known.
*/
char *Audio_MemberGetSDES(Audio_member_t m, const SDES_t type)
{
  struct sdes *sdes;

  for (sdes = m->sdes; sdes; sdes = sdes->next) {
    if (sdes->type == type) return sdes->value;
  }
  return "";
} /* Audio_MemberGetSDES */


/*
* Find member within session, based on 'src' source identifier.  If
* existing member has src=0, which occurs for RTP members, match on
* nsrc.  This can fail for multi-homed hosts or hosts with several
* senders; the same failure occurs in vat mode.  Add member via
* add_member() if not found.
*   s          in   session
*   type       in   control or data packet; flag bits
*                   MEMBER_DONT_ADD: don't add even if not found
*                   MEMBER_CONTROL:  control packet
*                   MEMBER_NODATA:   neither control nor data
*   *nsrc;     in   network source address
*   src        in   synchronization or content source;
*                   may be zero if RTP site is created locally
*                   from session_add_host().
*/
Audio_member_t member_find(Audio_session_t s, int type,
  tsap_t *nsrc, src_id_t src)
{
  Audio_member_t m = NULL;
  Tcl_HashEntry *entry;

  /* 
   * Check if src equals local member -> collision. 
   */
  if (s->local && src == s->local->src) {
    int new;

    debug(DEBUG_CNTL, "Collision %x", src);

    /* be nice and say BYE before changing identities */
    tx_packet_session(s, 0, P_bye);

    /* delete old local identifier and create new one */
    entry = Tcl_FindHashEntry(&s->member_table, (char *)(long)src);
    Tcl_DeleteHashEntry(entry);
    s->local->src = random32((int)s);
    entry = Tcl_CreateHashEntry(&s->member_table, 
               (char *)(long)s->local->src, &new);
    Tcl_SetHashValue(entry, s->local);

    /* reset statistics */
    s->packets[0] = 0;
    s->packets[1] = 0;
    s->bytes      = 0;
  }

  /*
   * Check member cache only for non-zero src, since zero src is a
   * locally created unicast RTP site.
   */
  if (src && s->cache && s->cache->src == src) {
    m = s->cache;
  }
  /* If not found in cache, search in hash table. */
  else {
    if (s->first) {
      entry = Tcl_FindHashEntry(&s->member_table, (char *)(long)src);
      if (entry) m = (Audio_member_t)Tcl_GetHashValue(entry);
    }

    /*
	 * If not in hash table, this might be an RTP source that matches
	 * (by network source address) an existing unicast member entered
	 * explicitly.  We simply fill in the src value and add it to the
	 * hash table.
     */
    if (src && !m && !session_ismulticast(s)) {
      int new;

      ALL_MEMBERS(s) {
        if (m->src == 0 && tsap_equ_a(nsrc, &m->nsrc[0])) {
          m->src = src;
          entry = Tcl_CreateHashEntry(&s->member_table, (char *)(long)src, &new);
          Tcl_SetHashValue(entry, m);
          break;
        }
      }
    }

    /*
	 * Don't add it unless we are supposed to.
     */
    if (!m && type & MEMBER_DONT_ADD) {
      return 0;
    }

    if (!m) m = member_add(s, type, nsrc, src);
    if (!m) return 0;

    /*
	 * Update cache only for data packets (control packets have no
	 * correlation).  Update non-zero only since zero indicates
	 * locally generated RTP site.
     */
    if (src && !(type & MEMBER_CONTROL))
      s->cache = m;
  }

  /* disable soft timeout */
  if (m->soft_to && m->id) {
    m->soft_to = FALSE;
    (*nevot.event)(s, m, Audio_alive, 1);
  }

  if (!(type & MEMBER_NODATA))
    gettimeofday(&m->last[type & MEMBER_CONTROL], 0);

  return m;
} /* member_find */

/*
* User-oriented member designation. Return "" if nothing found.
*/
char *Audio_MemberId(Audio_member_t m)
{
  if (m->sdes) return m->sdes->value;
  else return tsap_ntoa(&m->nsrc[1]);
  return "";
} /* Audio_MemberId */

/*
* Set listen status of session member 'm'.
* Return Audio_OK.
*/
Audio_errno_t Audio_MemberListen(Audio_member_t m, const int flag)
{
  Audio_session_t s = m->s;

  /* set receive mode only makes sense for non-local member */
  m->listen = flag;
  m->talkspurt = 1;
  if (s->local && m != s->local) {
    (*nevot.event)(s, m, Audio_receive, m->listen);
  }
  return Audio_OK;
} /* Audio_MemberListen */
