/*
ring_mix  --  mix audio into ring

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <stdlib.h>  /* free() */
#include "nevot.h"
#include "audio_generic.h"
#include "ring.h"
#include "mix.h"
#include "channels.h"



/*
* Mix audio chunk 'b' of size 'bytes' bytes and type 'ad' from member
* 'm' into ring 'ring' at position 'p' (samples).  'b' may be zero, in
* which case only the talking indication is adjusted.  The array
* 'mlist' contains 'sources' entries with Audio_member_t pointers
* indicating the contributing members.  Return Audio_OK or Audio
* error.
*/
Audio_errno_t ring_mix(Audio_member_t *mlist, int sources, 
  audio_descr_t *ad, char *b, int bytes, ring_t *ring, u_int32 p)
{
  u_int32 reach;         /* reach of audio (in samples) */
  char out_enc[8192];    /* output buffer for encoding conversion */
  char out_ch[8192];     /* output buffer for channel conversion */
  int i;
  Audio_member_t m;      /* SSRC */
  audio_descr_t ad_in = *ad;  /* changes as we convert */
  audio_descr_t *ad_out = &nevot.rx.ad;  /* destination */
  Audio_errno_t status;
  extern codec_t codec[];

  m = mlist[0];
  /* Can happen if s->local is zero (not sure why THIS can happen). */ 
  if (!m) return Audio_OK; 
  if (!m->listen) b = NULL;

  /*
   * Compute coverage before conversion since we won't necessarily
   * compute a correct byte count if b == NULL. 'reach' is extent
   * of audio beyond the insertion point 'p'.
   */
  m->covers = bytes2samples(&ad_in, bytes);
  reach = p + m->covers;

  if (b) {
    /* convert in-place to local audio format */
    if (ad_in.encoding != ad_out->encoding) {
      if (codec[ad->encoding].decoder) {
        status = (*codec[ad_in.encoding].decoder)(b, bytes, ad_out, 
                 out_enc, &bytes, m->decoder_state);
        if (status < 0) return Audio_EENC;
      }
      else return Audio_EENC;
      ad_in.encoding = ad_out->encoding;
      b = out_enc;
    }

    /* convert between stero and mono */
    if (ad_in.channels != ad_out->channels) {
      status = channels(b, bytes, &ad_in, out_ch, &bytes, ad_out, m->s->local_audio_channel);
      if (status < 0) return Audio_ECHANNELS;
      ad_in.channels = ad_out->channels;
      b = out_ch;
    }

    /* Flag error if wrong sampling rate. */
    if (ad_in.sample_rate != ad_out->sample_rate) return Audio_ERATE;

    /* if audio device closed, open it */
    if ((status = audio_open()) < 0) return status;
  }

  /*
   * Set talking indicator.
   */
  for (i = 0; i < sources; i++) talking_set(ring, mlist[i]);

  /*
   * Set last_sample for ring only if we actually copied audio into
   * the buffer.
   */
  if (b && reach > ring->last_sample)
    ring->last_sample = reach;

  m->stats.samples += m->covers;

  /* Determine audio reach for individual session member. */
  if (reach > m->last_sample)
    m->last_sample = reach;

  if (b) {
    switch(ad_out->encoding) {
    case AE_PCMU:
      ring_cpy(ring, p, m->covers, mix2_pcmu, b, ad_out->channels);
      break;
    case AE_L8:
      ring_cpy(ring, p, m->covers, mix2_l8,   b, ad_out->channels);
      break;
    case AE_L16:
      ring_cpy(ring, p, m->covers, mix2_l16,  b, ad_out->channels * 2);
      break;
    default:
      break;
    }
  }

  return Audio_OK;
} /* ring_mix */
