/*
rx  --  handle packets from network and play audio

We size the ring buffer to be a multiple of the current vox_size
so that playout, resetting and energy-measurement operations never cross
the physical end of the ring buffer.

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include "nevot.h"
#include "util.h"
#include "rx.h"          /* RX_t */
#include "control.h"
#include "ahw.h"
#include "des.h"         /* des_cbc_encrypt() */
#include "peakmeter.h"
#include "debug.h"
#include <math.h>        /* fabs() */
#include <stdlib.h>      /* abs(), malloc() */
#include <errno.h>
#include <time.h>        /* time() */
#include <sys/time.h>
#include <ansi.h>

#define MAX_SPEAKER 17  /* CSRC plus one SSRC */


extern int errno;
struct RX_t RX;

/* play file */
static struct {
  FILE *f;           /* file */
  int on;            /* enable recording */
} pf;

/*
* Bytes per sample, only for the decoded sizes.
*/
static int bps(audio_descr_t *ad)
{
  if (ad->encoding == AE_PCMU) return ad->channels;
  else return ad->channels * 2;
} /* bps */

/*
* memcpy() with reversed arguments.
*/
static void memcpy_r(char *s, char *d, u_int32 len)
{
  memcpy(d, s, len);
} /* memcpy_r */

/*
* Blind timing based on first packet in talkspurt and fixed (minimum)
* playout delay.
*/
static void talkspurt_fixed(struct delay_t *delay, sync_t *sync)
{
  sync->delay = sync->t + (unsigned)(delay->min * nevot.rx.ad.sample_rate) - sync->this.ts;
  sync->p     = sync->this.ts + sync->delay; 
  sync->slack = sync->p - sync->t;
  sync->d_est = MAX32U;  /* max. value to restart min() computation */
} /* talkspurt_fixed */


/*
* Beginning of a new talkspurt.
*
* We should (but currently do not) check whether the beginning of a new
* talkspurt gets mixed into the end of the previous talkspurt. This
* requires knowledge about the coverage of the previous talkspurt, which
* is not always available prior to audio decoding and mixing. Checking
* that the overlap is less than a packetization interval is easy, however.
*
* Use per-talkspurt minimum as constant delay estimate (propagation
* delay + transmission delays + clock difference) rather than average.
*
*/
static void talkspurt(struct delay_t *delay, sync_t *sync)
{
  u_int32 delay_old = sync->delay;
  u_int32 v;
  int min = delay->min * nevot.rx.ad.sample_rate, 
      max = delay->max * nevot.rx.ad.sample_rate;

  /*
   * Update playout delay.
   * If min >= max, we do blind timing, with the minimum delay as
   * as fixed value. For that, late_ts should be zero to avoid playing
   * every packet a fixed amount after its arrival.
   */
  if (min >= max) {
    talkspurt_fixed(delay, sync);
    return;
  }

  /* Adaptive playout delay. */
  v = (u_int32)(delay->multiplier * sync->d_var);
  if      (v > max) v = max;
  else if (v < min) v = min;
  sync->delay = v + sync->d_est;

  /* compute playout time of first packet in talkspurt */
  do {
    sync->p     = sync->this.ts + sync->delay; 
    sync->slack = sync->p - sync->t;

    if ((int)sync->slack >= 0) {
     /*
      * If new delay is smaller than old delay, need to assure that we
      * do not overlap old audio. We compare delay change to timestamp gap. 
      * Must be careful to handle case of sender restart.
      */ 
      if (sync->delay > delay_old) break;
      if ((int)(sync->p - sync->last_audio) >= 0) break;

      debug(DEBUG_TS, "overlap ts=%lu p=%lu last=%lu", 
        sync->this.ts, sync->p, sync->last_audio);
      sync->delay = sync->last_audio - sync->this.ts;
      if (sync->delay == 0) {
        printf("sync %lu %lu\n", sync->last_audio, sync->this.ts);
      }
    }
    /* 
     * If slack is still negative (i.e., packet late), play immediately.
     */
    else {
      sync->delay = sync->t - sync->this.ts;
    }
  } while (1);

  sync->d_est = MAX32U;  /* max. value to restart min() computation */
} /* talkspurt */

/*
* Update delay estimates, based on absolute timestamp difference 'd'
* and extra playout time 'slack'. Basing the variance estimate on
* slack variation cuts off the peaks that caused late packets.
*/
static void rx_stats(Audio_member_t m, struct delay_t *delay, 
  sync_t *sync, u_int32 d)
{
  /* sender/receiver clock difference: use per-talkspurt minimum */
  if (d < sync->d_est) sync->d_est = d;

  /*
   * If packet is late (can happen if no automatic talkspurt on late
   * packets), don't tally playout delay (we won't play.)
   */
  if ((int)sync->slack < 0) {
  }
  /* Playout delay is either fixed or below maximum allowed. */
  else if (delay->max <= delay->min ||
      (int)sync->slack <= delay->max * nevot.rx.ad.sample_rate) {
    m->stats.playout_delay += sync->slack;
    m->stats.pit++;
    m->stats.good++;
  }
  else talkspurt(&m->s->sp.delay, sync);
  
  /* estimate delay variation */
  sync->d_avg = delay->dvar_x * sync->slack + delay->dvar_y * sync->d_avg;
  sync->d_var = delay->dvar_x * fabs(sync->slack - sync->d_avg) 
     + delay->dvar_y * sync->d_var;
} /* rx_stats */

/*
* Show debugging info. type = 'T' (talkspurt) or ' '.
*/
static void rx_debug(int d, char type, int discard, int latest, sync_t *sync)
{
  debug(d, "%c%c%c %10u %10u %10u %10u %10u %10d %8.0f %8.0f %10u",
   type,
   discard ? '-' : '+',
   latest  ? 'l' : ' ',
   sync->this.ts, sync->this.seq, sync->t, sync->delay, sync->p, sync->slack,
   sync->d_avg, sync->d_var, sync->d_est);
} /* rx_debug */

/*
* Read packet from socket and place in circular buffer.
* The 'client' argument contains the 'Audio_session_t' pointer; if control,
* the lsb is set.
* Showing member talk mode is made difficult by the fact that arriving
* packet size may not be equal to audio interval.
*
* The delay 'd' of incoming packets may be negative due to clock
* differences.
*/
Notify_value rx(Notify_client client, int sock)
{
  Audio_session_t s;                 /* current session */
  Audio_member_t m = 0;              /* current member */
  static mbuf_t *mb;                 /* buffer for incoming audio packet */
  sync_t *sync;                      /* synchronization state */
  static tsap_t nsrc;                /* network source */
  struct {
    int count;                       /* number of sync/content sources */
    Audio_member_t mlist[MAX_SPEAKER];   /* content sources */
  } source;
  u_int d;                           /* offset between t and ts (samples) */
  int first = 0;                     /* first packet */
  int discard = 0;                   /* discard this packet */
  int latest = 0;                    /* this is the latest packet */
  int rbytes;                        /* bytes received */
  int mode;                          /* packet mode */
  Audio_errno_t status = 0;          /* error status */
  int control;                       /* control packet */
  struct hdr {   /* common header */
    unsigned int type:2;
    unsigned int dummy:6;
  } *hdr;

  /* initialization on first call */
  if (!mb) {
    mb = m_get("rx");
    tsap_alloc(&nsrc, T_sockaddr);
  }
  mb->m_off = MMINOFF(mb);

  /* control or data socket */
  s = (Audio_session_t)((long)client & ~0x1);
  control = (long)client & 0x1;

  rbytes = (*s->net.read)(sock, mtod(mb, void *), nevot.maxlen, &nsrc);
  if (rbytes >= nevot.maxlen) {
    debug(DEBUG_RX, "Received packet cut off at %d bytes", rbytes);
  } 

  if (rbytes < 0) {
	if (errno == ECONNRESET) {  /* other party left */
      rbytes = 0; /* treat like EOF */
	}
    /* not sure when/why these happen */
    else if (errno == ENETUNREACH) {
    }
    else if (errno == EHOSTUNREACH) {
    }
	else {
      (*nevot.event)(s, 0, Audio_error, Audio_ENETIN);
      return NOTIFY_DONE;
	}
  }
  mb->m_len = rbytes;

  /*
   * Decrypt whole data packet in place; need to do this at the very
   * beginning.  For data, bother only if we are listening to session.
   * For control packets, skip the random four bytes at beginning.
   * Problem: shouldn't do this for VAT packets, thus we kludge by
   * checking our sending protocol.
   */
  if (s->encrypt && (control || s->listen)) {
    /*des_cbc_encrypt(
      mtod(mb, des_cblock *),
      mtod(mb, des_cblock *),
      mb->m_len,
      *(des_key_schedule *)s->key,
      (des_cblock *)"\0\0\0\0\0\0\0\0",
      DES_DECRYPT);
    if (control && s->sp.rt_proto == R_rtp) m_adj(mb, 4);*/
  }

  /*
   * Determine real-time protocol being used and read header.
   */
  hdr = mtod(mb, struct hdr *);
  source.count = 0;
  mode = 0;
  m = 0;
  if (hdr->type == 0) {
    if (control)
      status = vat_read_control(s, mb, &nsrc, &m);
    else if (s->listen) {
      status = vat_read_data(s, mb, &nsrc, &source.count, source.mlist, &m);
    }
    else return NOTIFY_DONE;
  }
  else if (hdr->type == 2) {
    if (control)
      status = rtp_read_control(s, mb, &nsrc, &m);
    else if (s->listen) {
      status = rtp_read_data(s, mb, &nsrc, &source.count, source.mlist, &m);
    }
    else return NOTIFY_DONE;
  }
  else {
	debug(DEBUG_RX, "unknown header version %d from %s", hdr->type, 
	    tsap_ntoa(&nsrc));
    return NOTIFY_DONE;
  }

  if (status != Audio_OK) {
    if (!m) m = s->local;
    if (m) {
      if (status == Audio_EIAUDIO) m->stats.bad_fmt++;
      else m->stats.bad_header++;
    }
	debug(DEBUG_RX, "packet error (type %d) from %s: %s", hdr->type, 
	    tsap_ntoa(&nsrc), Audio_strerror(status));
    return NOTIFY_DONE;
  }

  /* Update network source. */
  tsap_cpy(&m->nsrc[control], &nsrc);

  /* Announce network source for very first data or control packet. */
  if ((control && !m->stats.control) || (!control && !m->stats.received)) {
    if (m->id) {
      char buf[80];

      sprintf(buf, "%s:%d:%d",
        tsap_ntoa(&m->nsrc[control]),
        tsap_getport(&m->nsrc[0]), tsap_getport(&m->nsrc[1])); 
      (*nevot.event)(s, m, Audio_from, buf);
    }
  }

  /* Update control packet average for session. */
  if (control) {
    rtcp_period(0, 0, 0, 0, rbytes, &s->control.avg, 0);
    m->stats.control++;
  }

  /* control packets all return here */
  if (!m || control) return NOTIFY_DONE;

  sync = &m->sync;
  m->active = 1;   /* reset when next control packet is sent */

  /* If audio encoding is invalid, discard now. */
  if (m->audio.encoding == AE_MAX) {
    m->stats.bad_fmt++;
    return NOTIFY_DONE;
  }

  /*
   * Count sequence number wrap-arounds; the first test works
   * only for sequence numbers with modulus less than 32 bits.
   */
  if (m->stats.received == 0) {
    sync->first = sync->this;
    first = TRUE;
  }

  /*
   * Update jitter estimate.
   * jitter = (R_j - R_i) - (S_j - S_i)
   * S: m->sync.this|prior.ts; R: sync->t (old), s->ring->next (new)
   */
  if (!first) {
    int jitter;  /* jitter sample */

    jitter = (s->ring->next - sync->t) - 
             (m->sync.this.ts - m->sync.prior.ts);

    m->stats.jitter += (abs(jitter) - (int)m->stats.jitter) / 16;
  }

  m->stats.received++;

  /* 
   * If the audio device is not open, we open it.
   * Should be harmless if we end up discarding the audio packet.
   */
  if ((first = audio_open()) < 0) {
    (*nevot.event)(s, 0, Audio_error, first);
    return NOTIFY_DONE;
  }
  sync->t = s->ring->next;

  /* XXX convert timestamp to rate of audio output format;
     needed if we do rate conversion */

  /* determine offset between sending and receiving timestamps */
  d = sync->t - sync->this.ts;

  /*
   * For demonstration purposes only, play out immediately.
   * Otherwise, compute provisional playout point.
   */
  if (s->sp.delay.immediate) {
    sync->p = sync->t;
  }
  else {
    sync->p = sync->this.ts + sync->delay;
  }
  sync->slack = sync->p - sync->t;

  /*
   * Recover from sudden jumps in sender timestamp. These can
   * occur if the sender restarts, e.g., after a crash. If the
   * playout instant falls outside the buffer window, we reset
   * the delay offset. Also called on very first audio packet.
   */
  if (first || abs(sync->slack) > s->ring->size) {
    sync->d_avg      = s->sp.delay.initial * nevot.rx.ad.sample_rate;
    sync->d_var      = sync->d_avg / s->sp.delay.multiplier;
    sync->delay      = 0;
    sync->d_est      = d;
    latest = TRUE;

    debug(DEBUG_RX, "delay.multiplier=%g ring.next=%u ring.size=%d", 
      s->sp.delay.multiplier, s->ring->next, s->ring->size);

    /* header for timing reports */
    debug(DEBUG_RX, "%c%c%c %10s %10s %10s %10s %10s %10s %8s %8s %10s",
      'T', '+', 'l', "timestamp", "seqno", "t", "delay", "p", "slack", "d_avg",
      "d_var", "d_est");

    talkspurt(&s->sp.delay, sync);
  } 
  /*
   * Check for duplicate packets. Only packets duplicating immediate
   * predecessor are detected.
   */
  else if (sync->this.ts == sync->prior.ts && sync->this.seq == sync->prior.seq) {
    m->stats.dup++;
    discard = TRUE;
  }
  /*
   * Packet is the latest if it contains the highest sequence #.
   */
  else if (sync->this.seq == sync->max_seq) latest = TRUE;

  /* Is packet late (playout instant has passed)? */
  if ((int)sync->slack < 0) {
    m->stats.late++;
    if (latest) {
      if (s->sp.delay.late_ts) {
        talkspurt(&s->sp.delay, sync);
      }
      else {
        discard = TRUE;
      }
    }
    /*
     * If not latest, discard.
     */
    else if (m->talkspurt) {
      talkspurt(&s->sp.delay, sync);
    }
    else {
      discard = TRUE;
    }
  }
  /*
   * Else, if latest packet and talkspurt, start a new talkspurt.
   */
  else if (latest) {
    if (m->talkspurt || s->force_ts)
      talkspurt(&s->sp.delay, sync);
    sync->last_audio = sync->p;
  }

  /* count talkspurts */
  if (m->talkspurt) {
    rx_debug(DEBUG_TS|DEBUG_RX, 'T', discard, latest, sync);
    m->stats.marker++;
    m->talkspurt = 0;
  }
  else {
    rx_debug(DEBUG_RX, ' ', discard, latest, sync);
  }

  /* tally playout delay and update variance estimate */
  rx_stats(m, &s->sp.delay, sync, d); 

  if (!discard) {
    /* mix incoming audio into ring buffer */
    status = ring_mix(&m, 1, &m->audio, mtod(mb, char *), mb->m_len, 
                      s->ring, sync->p);
    if (status != Audio_OK)
      (*nevot.event)(s, m, Audio_error, status);
  }

  s->force_ts = FALSE;
  m->sync.prior = m->sync.this;

  /* play the input data as soon as they are received */
  if(s->listen&&!control){   
    tx(0,sock);
  }
  return NOTIFY_DONE;
} /* rx */

Audio_errno_t rx_file_stop()
{
  if (pf.f) fclose(pf.f);
  pf.on = FALSE;
  return Audio_OK; 
} /* rx_file_stop */

Audio_errno_t rx_file_pause(int flag)
{
  pf.on = !flag;
  return Audio_OK; 
} /* rx_file_pause */
/*
* Configure recording file with file descriptor 'fd'.
* File 'info' and * name 'fn'.
*/
Audio_errno_t rx_file(int fd, char *info, char *fn)
{
  pf.f  = af_record(fd, fn, &nevot.rx.ad, info);
  pf.on = TRUE;
  return Audio_OK; 
} /* rx_config_file */

static void play_silence(char *dst, char *src, int len)
{
  audio_silence(dst, len, nevot.rx.ad.encoding);
} /* play_silence */


/*
* Play next audio segment from ring buffer, possibly mixing in local
* sidetone. If 'mute', don't play.
*/
void play_local(void *sidetone, audio_descr_t *ad, int sidetone_bytes,
  int monitor, int mute)
{
  ring_t *r;                 /* current ring */
  static audio_stats_t as;   /* audio statistics */
  static ring_t *play_ring;  /* for audio playback without session */
  static int underflow;

  /* determine ring buffer to use; with and without existing session */
  if (nevot.first) {
    r = nevot.first->ring;
  }
  else {
    if (!play_ring) play_ring = ring_create(nevot.rx.ad.encoding);
    r = play_ring;
  }

  /*
   * If enabled, mix sidetone into buffer. Disable sidetone in
   * case we are also recording.
   */
  if (monitor) {
    ring_mix(&nevot.first->local, 1, ad, sidetone, sidetone_bytes, r, r->next);
    sidetone = (void *)0;
  }

  /* Declare members that are no longer talking as inactive. */
  talking_check(r);

  /* 
   * Compute audio statistics and play to audio device unless there is
   * no audio in buffer. After copying to audio device, clear buffer.
   * To avoid starvation and clicks during short pauses, allow about
   * a second worth of silence.
   */
  if (ring_left(r) > -8000) {
    ring_action(r, r->next, nevot.rx.samples, (ring_action_t *)audio_stats,
       (void *)&nevot.rx.ad, (void *)&as, bps(&nevot.rx.ad));
    if (!mute) {
      ring_action(r, r->next, nevot.rx.samples, (ring_action_t *)ahw_write,
        0, 0, bps(&nevot.rx.ad));

      /* 
       * Every so often, check queue occupancy. If underflow error,
       * write audio block twice.
       */
      if (++underflow > 50) {
        int q = ahw_write_queue(&underflow);
        if (underflow) {
          ring_action(r, r->next, nevot.rx.samples, (ring_action_t *)ahw_write,
            0, 0, bps(&nevot.rx.ad));
          debug(DEBUG_AUDIO, "audio underflow (%d)", q);
        }
      }
    }
    peakmeter(RX.pm, nevot.rx.ad.encoding, &as);
  }
  else {
    as.c[0].peak = as.c[1].peak = 0;
    peakmeter(RX.pm, nevot.rx.ad.encoding, &as);
  }

  /* 
   * Mix local audio if we are recording; if we are monitoring, sidetone
   *   has been set to zero earlier.
   */
  if (pf.on) {
    ring_mix(&nevot.first->local, 1, ad, sidetone, sidetone_bytes, r, r->next);
    ring_action(r, r->next, nevot.rx.samples, (ring_action_t *)fwrite,
      (void *)1, pf.f, bps(&nevot.rx.ad));
  }

  /*
   * Clear audio buffer again.
   */
  if (ring_left(r) > -8000) {
    ring_cpy(r, r->next, nevot.rx.samples, play_silence, 0, bps(&nevot.rx.ad));
  }

  /* increment playout ring pointer */
  r->next += nevot.rx.samples;
} /* play_local */

/*
* For each session, get current packet and send it off to all sessions not
* sharing the same ring buffer.  Sooner or later, we need to copy the
* audio from the ring buffer into an mbuf.  We do it here.
*/
void play_gateway(void)
{
  Audio_session_t s;
  ring_t *r;
  mbuf_t *mb;
  int talk_save;

  ALL_SESSIONS(s) {
    r = s->ring;

    /* send to all other sessions */
    talk_save = s->talk;
    mb = m_get("pg");
    mb->m_len = nevot.rx.bytes;
    mb->m_ad  = nevot.rx.ad;
    mb->m_type |= P_csrc;  /* identify contributors */
    ring_cpy(r, r->next, mb->m_len, memcpy_r, mtod(mb, char *), 1);

    /* tx_packet avoids sending to self; also frees mbuf */
    tx_packet(mb, r);
    r->next += nevot.rx.samples;
    s->talk = talk_save;
  }
} /* play_gateway */

/*
* Close those members that have not received either a control message or
* an audio packet within the specified timeout interval. Audio timeout
* is disabled if the member is not being listened to (this allows remote
* members that are only listening).  Control timeout is disabled for members
* that have never sent a control message. As the number of members
* grows, the function is called with increasing period, limiting the
* search load.
*
* Alternative 1: Instead of a periodic linear search through all members,
* we could keep an LRU least of members, rearranged each time a member
* receives a packet.  We then only need to check a probably small
* subset of members, until the last packet arrival interval is smaller
* than the timeout interval.  This requires per-packet LRU list
* insertion (at head) and deletion. 
*
* Alternative 2: We could keep a timer for each member that goes off
* after the soft/hard timeout.  Each packet arrival restarts the
* timer.  The cost depends on the efficiency of location the timer
* data structure and requires more work than the LRU chain (deletion
* and insertion in proper location of timer chain, which is going to
* require at least O(log N) comparisons), but has the advantage that
* no periodic checking is required. 
*/
void rx_periodic(Audio_session_t s)
{
  Audio_member_t m, m_next; /* need to save since we destroy current one */
  int i;                /* timeout index (soft/hard) */
  time_t expire;        /* expiration time (seconds) */
  time_t t = time(0);   /* current time */

  /* i=0: soft (grey); i=1: hard (remove) */
  for (i = 0; i < 2; i++) {
    expire = t - (s->control.period * RX.timeout[i]);

    for (m = s->first; m; m = m_next) {
      m_next = m->next;
      if (m != s->local &&     /* don't time-out local member */
          m->last[0].tv_sec < expire &&
          m->last[1].tv_sec < expire &&
          m->created < expire &&
          /* don't time out sites we haven't heard from yet */
          (m->last[0].tv_sec > 0 || m->last[1].tv_sec > 0)
         ) {

        if (i == 0) {  /* soft timeout */
          if (!m->soft_to) {
            m->soft_to = TRUE;
            if (m->id) (*nevot.event)(s, m, Audio_alive, 0);
            debug(DEBUG_AUDIO, "timeout%d %s: t=%u; expire=%u; audio=%u; control=%u",
              i, tsap_ntoa(&m->nsrc[0]), t, expire,
              m->last[0].tv_sec, m->last[1].tv_sec);
          }
        }
        else Audio_MemberDelete(m, "timeout");
      }
    } /* for s */
  } /* for i */
} /* rx_periodic */   

/*
* Initialize RX structure to resaonable defaults.
*/
void rx_init(void)
{
  RX.timeout[0]            = 5;
  RX.timeout[1]            = 20;
  audio_chunk();
  RX.pm = peakmeter_init(1, audio_vu, audio_vu_max);

  nevot.rx.samples         = 160;
  nevot.rx.bytes           = 160; 
  /*
   * Initialize to invalid values to force proper audio
   * hardware initialization through audio_change().
   */
  nevot.rx.ad.encoding     = 0;
  nevot.rx.ad.channels     = 0;
  nevot.rx.ad.sample_rate  = 0;
} /* rx_init */
