/*
tx_packet  ---  transmit single packet on network

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include <sys/types.h>
#include "nevot.h"
#include "msghdr.h"
#include "channels.h"
#include "debug.h"
#include "des.h"           /* des_cbc_encrypt() */
#include "audio_generic.h"
#include <stdlib.h>        /* free() */
#include <assert.h>
#include <ansi.h>


extern codec_t codec[];    /* codec description */

/* encoder state and output buffer for each encoding */
static struct {
  mbuf_t *m;         /* packet */
  void *state;       /* encoder state */
} cs[AE_MAX];


/*
* Initialize packet transmission (coder, etc.)
* Can be used after change of relevant parameters (packetization interval).
*/
void tx_packet_init(void)
{
  audio_encoding_t c;

  for (c = 0; c < AE_MAX; c++) {
    if (cs[c].state) {
      free(cs[c].state);
      cs[c].state = 0;
    }
    if (codec[c].encoder) {
      cs[c].state = (void *)(*codec[c].init)(0, nevot.audio.tick);
      if (cs[c].m) m_free(cs[c].m, "tpi");
      cs[c].m = 0;
    }
  }
} /* tx_packet_init */


/*
* Clean up codecs.
*/
void tx_packet_close(void)
{
  audio_encoding_t c;

  for (c = 0; c < AE_MAX; c++) {
    if (cs[c].state) free(cs[c].state);
  }
} /* tx_packet_close */


/*
* Add mbuf 'mb' to the list of mbufs belonging to session 's'.
*/
static void tx_session_attach(Audio_session_t s, mbuf_t *mb)
{
  if (mb) {
    mb->m_ref++;
    mb->m_next = NULL;
    if (s->b_tail) s->b_tail->m_next = mb;
    else s->b_head = mb;
    s->b_tail = mb;
    s->block++;
  }
} /* tx_session_attach */


/*
* Sent message 'mb' to single session member.  Lower-level routine
* will ignore call if member has no valid 'control/data' record,
* returning zero. Don't send data to members with src=0, as those are
* RTP entities that have been created locally. If they exist, they
* will rep
*/
static void tx_member(Audio_session_t s, Audio_member_t m,
  int control, mbuf_t *mb)
{
  int sent;  /* bytes sent */

  control &= 1;
  sent = (s->net.write)(control ? s->net.control : s->net.data, mb, 
    &m->ndst[control]);
  if (sent > 0) m->stats.sent++;
  else if (sent < 0) (*nevot.event)(s, m, Audio_error, Audio_ENETOUT);
} /* tx_member */


/*
* Transmit audio packet for a single session.  If s->b_head is zero,
* this is a control packet.  ring_skip denotes ring where audio is
* coming from.  Control options describe audio packet.  Need id for
* labeling control packets without audio.  Return size of packet sent.
*/
int tx_packet_session(Audio_session_t s, ring_t *ring_skip, int packet_type)
{
  int control;       /* control packet */
  mbuf_t *mb;        /* mbuf chain */
  Audio_member_t m;  /* members within session */
  int i;
  int size;

  /* save mbuf state of first and last mbuf prior to adding RTP/VAT header */
  mbuf_t *mdata[2];
  unsigned long m_off[2];
  short m_len[2];

  if (!s->net.write) return 0;   /* nothing to write to */

  /* distinguish between data and control packets */
  control = ((packet_type & 0xfff0) != 0);
  if (control) {
    mb = (*s->rt.write_control)(s, packet_type, 0);
  }
  else {
    /* Remember payload in mbuf, before adding headers. */
    if (s->b_head) {
      mdata[0] = s->b_head;
      mdata[1] = s->b_tail;
      for (i = 0; i < 2; i++) {
        m_len[i] = mdata[i]->m_len;
        m_off[i] = mdata[i]->m_off;
      }
    }
    mb = (*s->rt.write_data)(s, ring_skip, packet_type, s->b_head);
    if (packet_type & P_bot) debug(DEBUG_TX, "tx bot ts=%u", s->b_head->m_timestamp);
    s->seq++;
    s->local->active = 2;  /* reset when sending next control packet */
    /* clear chain of transmitted mbufs */
    s->b_head = s->b_tail = NULL;
    s->block  = 0;
  }
  size = mb->m_len;
  s->packets[control]++;

  /* Encrypt data and control packets here. */
  if (s->encrypt) {
    /*mbuf_t *m, *me = m_get("enc");

    for (m = mb; m; m = m->m_next) { 
      des_cbc_encrypt(mtod(m, des_cblock *),
        mtod_end(me, des_cblock *),
        m->m_len,
        *(des_key_schedule *)s->key,
        (des_cblock *)(m == mb ? "\0\0\0\0\0\0\0\0" : mtod_end(me, char *) - 8),
        DES_ENCRYPT);
      me->m_len += m->m_len;
    }
    m_freem(mb, "enc");
    mb = me;*/
  }

  if (session_ismulticast(s)) tx_member(s, s->local, control, mb);
  else ALL_MEMBERS(s) {
    if (m != s->local) tx_member(s, m, control, mb);
  }

  /*
   * Restore first and last data mbuf to state without header, in case
   * we re-use these mbufs for some other session. This removes the 
   * RTP/vat header from the first and encryption padding from the
   * last mbuf in a chain.
   */
  if (!control) {
    for (i = 0; i < 2; i++) {
      mdata[i]->m_len = m_len[i];
      mdata[i]->m_off = m_off[i];
    }
  }

  m_freem(mb, "tps");  /* free all mbufs in chain */
  return size;
} /* tx_packet_session */

/*
* Transmit packet to one or more sessions.  Since the same compression
* algorithm is unlikely to be used for several active sessions, we
* currently don't bother keeping track of encodings already done. This
* would be easy to add by maintaining chains of sessions with the same
* encoding.
* 
* mbuf management: Each time we attach an mbuf to a chain, we
* increment its reference count. If the original mbuf is not used
* directly (since only transcoded versions are transmitted), its
* reference count will be zero and thus will be deleted immediately.
* For the other mbufs, they'll be deleted after transmission.
*/
void tx_packet(mbuf_t *m_orig, void *ring_skip)
{
  u_int32 timestamp;    /* mbuf timestamp */
  short type;           /* mbuf type */
  mbuf_t *m_enc = 0;    /* encoded data for current session */
  mbuf_t *ma;           /* converted audio data */
  Audio_session_t s;    /* current session */
  int status;           /* encoder status */
  int len;              /* output length */

  /* save timestamp and type (P_bot, P_eot) into final mbuf */
  type      = m_orig->m_type;
  timestamp = m_orig->m_timestamp;

  ma = m_orig;
  ALL_SESSIONS(s) {
    int c = s->audio.encoding;

    /* don't bother with session we won't send to */
    if (!s->talk || !s->net.write || (void *)s->ring == ring_skip) continue;

    /* net mutes mike */
    if (s->m.net_priority && talking_net(s->ring)) continue;

    /* Check if we have an appropriately formatted mbuf already. */
    if ((s->audio.encoding != ma->m_ad.encoding) ||
        (s->audio.channels != ma->m_ad.channels) ||
        (s->audio.sample_rate != ma->m_ad.sample_rate)) {
      ma = m_orig;  /* start fresh */

#ifdef RATE_CONVERSION
      if (s->audio.sample_rate != ma->m_ad.sample_rate) {
        m_enc = m_get("txr");
        m_enc->m_ad = ma->m_ad;
        status = rate(mtod(ma, char *), ma->m_len, &m_enc->m_ad,
                      mtod(m_enc, char *), &len, &s->audio,
                      &s->rate_state);
        m_enc->m_ad.sample_rate = s->audio.sample_rate;
        m_enc->m_len            = len;
        if (status < 0) {
          (*nevot.event)(s, 0, Audio_error, Audio_ERATE);
          m_free(m_enc, "txr.");
          continue;
        }
        if (ma != m_orig) m_free(ma, "txr");
        ma = m_enc;
        ma->m_len = len;
      } /* if sample_rate */
#endif

      if (s->audio.channels != ma->m_ad.channels) {
        m_enc = m_get("txc");
        m_enc->m_ad = ma->m_ad;
        status = channels(mtod(ma, char *), ma->m_len, &m_enc->m_ad, 
                          mtod(m_enc, char *), &len, &s->audio, 
                          0);
        m_enc->m_ad.channels = s->audio.channels;
        m_enc->m_len         = len;
        if (status < 0) {
          (*nevot.event)(s, 0, Audio_error, Audio_ECHANNELS, ma->m_ad.channels);
          m_free(m_enc, "txc.");
          continue;
        }
        if (ma != m_orig) m_free(ma, "txc");
        ma = m_enc;
      } /* if channels */

      if (s->audio.encoding != ma->m_ad.encoding) {
        m_enc = m_get("txe");
        m_enc->m_ad = ma->m_ad;
        status = (*codec[c].encoder)(mtod(ma, char *), ma->m_len, &ma->m_ad,
                 mtod(m_enc, u_int8 *), &len, cs[c].state, 
                 s->block == 0);
        m_enc->m_ad.encoding = s->audio.encoding;
        m_enc->m_len         = len;
        if (status < 0) {
          (*nevot.event)(s, 0, Audio_error, Audio_EENC);
          m_free(m_enc, "txe.");
          continue;
        }
        if (ma != m_orig) m_free(ma, "txe");
        ma = m_enc;
      } /* if encoding */
    } /* if != */

    ma->m_timestamp = timestamp;
    ma->m_type      = type;
    s->bytes += ma->m_len;
    tx_session_attach(s, ma);
  } /* ALL_SESSIONS */

  /* transmit complete messages */
  ALL_SESSIONS(s) {
    if ((void *)s->ring == ring_skip) continue;

    /*
     * Prevent crashes if there's no audio.
     */
    if (!s->b_head) continue;

    /* Transmit packets if we reached block size or the new mbuf
     * is the end of a talkspurt.
     * If encoding is changed when s->block is > 1 and the new bpm
     * is 1, the chain may never be transmitted. Thus, we have to
     * let session_encoding_change() clear the current chain or use the
     * test 's->block >= bpm'; in the latter case, mbufs with
     * different encodings get mixed into one packet.
     */
    if ((s->block >= s->bpm) || (s->block && (ma->m_type & P_eot))) {
      tx_packet_session(s, ring_skip, s->b_head->m_type);
    }
  }

  /*
   * Discard original (audio input) mbuf if it hasn't been put on
   * a linked list. If it's on one list, it will have ref=0, and
   * been marked as freed.
   */
  if (m_orig->m_ref == 0 && m_orig->m_type != MT_FREE)
    m_free(m_orig, "tp2");
} /* tx_packet */
