/*
tx  --  audio from local audio device to network

To allow for multiple voice protocols, the protocol header is filled in
as late as possible. Since the encoding copies the voice data, we can
insert the time stamp only on a per-session basis. 

Copyright 1991-1995 by Henning Schulzrinne; all rights reserved
*/

#include "ahw.h"
#include "nevot.h"
#include "util.h"        /* gettimeofday_double() */
#include "tx.h"          /* TX_t */
#include "cache.h"
#include "peakmeter.h"
#include "debug.h"
#include <sys/socket.h>
#include <stdlib.h>      /* drand48() [Sun] */
#include <stdio.h>       /* fread() */
#include <math.h>        /* drand48() */
#include <ansi.h>        /* drand48() */


struct TX_t TX;

Audio_errno_t tx_file_stop(void)
{
  if (TX.f.f) {
    fclose(TX.f.f);
    TX.f.f  = 0;
    TX.f.on = 0;
  }
  return Audio_OK;
} /* tx_file_stop */

/*
* Pause (flag=1)/unpause (flag=0)
*/
Audio_errno_t tx_file_pause(int flag)
{
  TX.f.on = !flag && TX.f.f;
  return Audio_OK;
} /* tx_file_pause */

Audio_errno_t tx_file_loop(int flag)
{
  TX.f.loop = flag;
  return Audio_OK;
} /* tx_file_loop */

Audio_errno_t tx_file_monitor(int flag)
{
  TX.f.monitor = flag;
  return Audio_OK;
} /* tx_file_monitor */

/*
* Configuration for playing audio files to remote members.
*/
Audio_errno_t tx_file(int fd, audio_file_t *af)
{
  Audio_errno_t status;

  /* Don't play if there is no open session (i.e., no ring buffer). */
  if (!nevot.first) return Audio_ENOSESSION;

  /* start new play */
  TX.f.f     = af_open(fd, "r", &TX.f.af);
  TX.f.bytes = audio_seconds_to_bytes(&TX.f.af.ad, nevot.audio.tick);
  TX.f.on    = TRUE;
  tx_packet_init();   /* reset encoders */
  if ((status = audio_open()) < 0) return status;

  TX.monitor = TX.f.f ? TX.f.monitor : TX.in.monitor;
  if (af) *af = TX.f.af;
  return Audio_OK;
} /* tx_file */

/*
* If no audio output upcoming (within release bytes) and recently,
* relinquish the audio output device; delay to avoid thrashing
* during short silences; possibly only relinquish if some other
* process really wants it.
*/
static void tx_periodic_audio(Audio_session_t s)
{
  /* check if others might be idle, too */
  if (ring_left(s->ring) < -TX.release * nevot.tx.ad.sample_rate)
    audio_close_idle();
} /* tx_periodic_audio */


/*
* Read one packet's worth of audio data from the current audio input
* file. If end of file is reached, either close the file or reset to
* start of file when looping. The last packet is filled with silence.
*
* bytes_need  I   number of bytes to read
* ad          O   description of audio data to be read
* vox         O   buffer for data
*/
static void tx_from_audio_file(int bytes_need, audio_descr_t *ad, char *vox)
{
  int bytes_read;    /* number of bytes read most recently */

  while (bytes_need > 0) {
    bytes_read = fread(vox, sizeof(char), bytes_need, TX.f.f);

    bytes_need -= bytes_read;
    vox = (char *)vox + bytes_read;
    if (bytes_need > 0) { 
      if (TX.f.loop) af_reset(TX.f.f, &TX.f.af);
      else {
        tx_file_stop();
        break;
      }
    }
  }
} /* tx_from_audio_file */

/*
* If not silent, transmit local audio (from microphone or file).  Note
* that audio from file may have a different format than the currently
* selected workstation audio input, but the frame duration in ms must be
* the same.  Also, invoke play_local() routine to play audio on local
* speaker.
*
* Can't use rand() for random number since it returns 32 bit numbers on
* BSD systems and 16 bit numbers on SysV systems.
*/
static void tx_local_audio(mbuf_t *mb)
{
  static int spurt_after;   /* packets transmitted after talk spurt */
  boolean net_talking = 0;  /* the net is talking */
  boolean talking;          /* we're talking */
  static audio_stats_t as;  /* statistics of current packet */
  char *vox;                /* audio data */

  vox = mtod(mb, char *);
  talking = nevot.audio.talk;

  /* read from audio file if enabled; set audio descriptor for mbuf */ 
  if (TX.f.on && TX.f.f) {
    mb->m_ad  = TX.f.af.ad;
    mb->m_len = TX.f.bytes;
    tx_from_audio_file(TX.f.bytes, &mb->m_ad, vox);
  }
  /* read from audio device */
  else {
    mb->m_ad  = nevot.tx.ad;
    mb->m_len = nevot.tx.bytes;
  }

  /*
   * Determine if there's audio in the playout buffer.
   * Currently, we don't handle the case that while there's audio,
   * the about-to-be-played part is silent.
   */
  if (nevot.first) {
    net_talking = ring_left(nevot.first->ring) > 0;
  }

  if (talking) {
    audio_stats(vox, mb->m_len, &mb->m_ad, &as);
    /*
     * Silent packet: if not trailer, insert into cache for beginning
     * of next talkspurt. If cache is full, only then will we have
     * a true talkpurt with a gap.
     */
    if (sd(TX.sd, &as, TX.echo_supp && net_talking > 0)) {
      TX.silent++;
      spurt_after--;
      talking = (spurt_after >= 0);
      if (!talking) {
        cache_write(mb, TX.spurt[0], &TX.bot);
        mb->m_type |= P_eot;  /* mark end of talkspurt */
      }
    }
    else {  /* valid audio packet (first or middle of talkspurt) */
      mbuf_t *mc;

      /* new talkspurt; transmit pre-talkspurt packets */
      while ((mc = cache_read())) {
        if (TX.bot) {
          mc->m_type |= P_bot;
          TX.bot = FALSE;
        }
        tx_packet(mc, (void *)0);
      }

      spurt_after = TX.spurt[1];
      talking = TRUE;
    }
    debug(DEBUG_SD, "sd %d %f", talking, as.c[0].peak);
  }
  /* If we don't transmit, discard audio buffer. */
  else {
    m_free(mb, "tla");
  }

  play_local(vox, &mb->m_ad, mb->m_len,
    TX.monitor && nevot.audio.talk,
    talking && nevot.first->m.local_priority);

  /* Transmit audio data and mark us as talking. */
  if (talking) {
    Audio_session_t s;

    if (!TX.f.on) {
      int inc = agc(TX.agc, &as);
      if (inc) ahw_set(0, AS_gain_step, inc);
    }
    if (TX.bot) {
      mb->m_type |= P_bot;
      TX.bot = FALSE;
    }
    tx_packet(mb, (void *)0);
    /* talker indication */
    ALL_SESSIONS(s) {
      ring_mix(&s->local, 1, &nevot.tx.ad, 0, mb->m_len, 
               s->ring, s->ring->next);
    }
  }

  /*
   * Display outgoing VU meter value, even if silent.
   */
  if (nevot.audio.talk) peakmeter(TX.pm, mb->m_ad.encoding, &as);
} /* tx_local_audio */

/*
* Functions to be performed periodically for each session:
* - check for timed-out members
* - transmit control packets
* - check for audio timeout
* client: session to check
*/
Notify_value tx_periodic(Notify_client client)
{
  Audio_session_t s = (Audio_session_t)client;

  s->control.last = tx_packet_session(s, (ring_t *)0, P_sdes | P_traffic);
  tx_periodic_audio(s);
  rx_periodic(s);

  s->control.period = rtcp_period(s->members, 
    s->senders + (s->local->active ? 1 : 0), 
    s->control.bw, s->local->active, s->control.last, &s->control.avg, 
    s->control.avg == 0.);
  s->senders = 0;

  notify_set_periodic_func((Notify_client)s, tx_periodic, 
    (int)(s->control.period * (drand48() + 0.5) * 1000));
  return NOTIFY_DONE;
} /* tx_periodic */

/*
* Called when (re)starting audio input.  To achieve some timestamp
* continuity, we add the time since the previous start to the initial
* timestamp, subtracting the current sample clock.  This fails if the
* outgoing audio sampling rate changes.
*/
void tx_start(void)
{
  Audio_session_t s;
  static double started = 0.;  /* very first invocation */
  double t;

  TX.bot = 1;

  /* reset talking indicator */
  t = gettimeofday_double();
  ALL_SESSIONS(s) {
    s->ring->last_sample = 0;
    talking_clear(s);
    if (started) {
      s->initial_ts += (t - started) * s->audio.sample_rate;
    }
  }
  TX.samples = 0;
  started = gettimeofday_double();
} /* tx_start */

/*
* Return accumulated sample count.
*/
u_int32 tx_samples(void)
{
  return TX.samples;
} /* tx_samples */

/*
* Take a buffer off the audio-device (fd) and transmit.
* Read as often as necessary to fill the packet. This routine has
* to be reentrant; it can happen that in calling tx_local_audio(),
* a 'send' is executed, during which time another audio buffer
* becomes ready, triggering a second instance of tx(), which
* reuses the same (static) mbuf mb. The same mbuf will then be free'd
* twice later.
*
* If the 'client' parameter is zero, data is read from a file. If not,
* we are on a system where tx() is called periodically by a timer
* (DEC AudioFile, for example).
*/
Notify_value tx(Notify_client client, int fd)
{
  static mbuf_t *mb;           /* pointer to buffer for packet */
  mbuf_t *mt;                  /* temporary mbuf */
  int bytes_new;               /* bytes just read */

  /*
   * Make sure we do this only once per cycle; don't set to zero just in
   * case ahw_read returns zero.
   */
  if (!mb) {
    mb = m_get("tr");
    mb->m_off += 200;  /* allow room for max. header size */
  }
  
  /*
   * If encoding changes on us (and thus nevot.tx.bytes),
   * we need to reset the counter to avoid spurious read errors.
   */
  /*  if (mb->m_len > nevot.tx.bytes) mb->m_len = 0;

  bytes_new = ahw_read(mtod_end(mb, void *), nevot.tx.bytes - mb->m_len);
  if (bytes_new < 0) {
    (*nevot.event)(nevot.first, 0, Audio_error, Audio_ERROR);
  }
  else {
    mb->m_len += bytes_new;
  }
  debug(DEBUG_TX, "tx read=%d", bytes_new);*/

  /*
   * If this is the timer version, schedule next invocation.
   */
  if ((int)client) {
    int left, dummy;
   
    left = ahw_read_queue(&dummy);
    /* Convert to milliseconds. */
    left = ((double)left / nevot.rx.ad.sample_rate) * 1000;

    /* Make sure new timer is not smaller than 10 ms. */
    if (left >= 10) left = 10;
    notify_set_periodic_func((Notify_client)1, (Notify_func)tx, 20 - left);
  }

  /*  if (mb->m_len < nevot.tx.bytes) return NOTIFY_DONE;*/

  /* after buffer has been filled */
  mb->m_timestamp = TX.samples;

  mt = mb;
  mb = 0;  /* get a new buffer next time; just in case mb gets free'd in tx_local_audio */

  if (nevot.role == R_gateway) {
    m_free(mt, "tr");
    play_gateway();
  }
  else {
    tx_local_audio(mt);
  }
  TX.samples += nevot.tx.samples;
  return NOTIFY_DONE;
} /* tx */

/*
* Default values for TX structure.
*/
void tx_init(void)
{
  TX.spurt[0]          = 1;
  TX.spurt[1]          = 3;
  TX.release           = 4;
  TX.sd  = sd_init();
  TX.agc = agc_init();
  TX.pm  = peakmeter_init(0, audio_vu, audio_vu_max);

  srand48(random32(0));  /* initialize random number generator */

  nevot.tx.samples         = 160;
  nevot.tx.bytes           = 160; 
  /*
   * Initialize to invalid values to force audio_change to
   * set the correct values.
   */
  nevot.tx.ad.encoding     = 0;
  nevot.tx.ad.channels     = 0;
  nevot.tx.ad.sample_rate  = 0;
} /* tx_init */
