/*
vat  --  routines for handling vat-format packets (audio and control).

vat is the native packet format for Van Jacobson's and Steve McCanne's
vat tool.

vat audio formats are:
     bytes   segments   audio  ts inc.
PCM  160     1          20 ms  160
PCM2 320     2          40 ms  320
DVI   84     1          20 ms  160
DVI4         4          80 ms
GSM  132     4          80 ms  640
LPC1         1          20 ms  160
LPC4         4          80 ms  640

Vat apparently does not like incomplete multi-frame packets (e.g., GSM
with only 3 20-ms frames instead of 4).  Since correcting for that is
rather difficult without contortions in a multi-session environment, we
punt.

---
The mapping between vat control and RTP/nevot control types is:

vat_id      P_sdes (local speaker)
vat_done    P_bye
vat_idlist  P_sdes_all (all contributors)

Copyright 1994 by Henning Schulzrinne; all rights reserved
*/

#include "nevot.h"
#include "audio_generic.h"
#include "vat.h"
#include "vat_session.h"
#include "ring.h"
#include "bit.h"         /* ROUND() */
#include "util.h"        /* strcpy_pad() */
#include "member.h"
#include "tsap.h"
#include "debug.h"       /* debug() */
#include <stdlib.h>      /* free() */
#include <assert.h>
#include <netinet/in.h>  /* htons() */

#define MAX_PT 32
#define VAT_SEQ_MOD  4294967296ULL
#define MAX_MISORDER 100
#define MAX_DROPOUT  3000


audio_descr_t vat_ad[MAX_PT];
static int vat_cpp[MAX_PT];   /* timestamp (clock) increment per packet */

/*
* Update cycle count based on new sequence number.
*
* small positive delta (0 < sdelta < MAX_DROPOUT):
*   new < old: wrap
*   new > old: 
*   both: just update
* small negative delta (0 < -sdelta < MAX_MISORDER):
*   misordered: do nothing
* big |delta| [or MAX_DROPOUT < udelta < RTP_SEQ_MOD - MAX_MISORDER]
*   check if bad
*/
static int update_seq(sync_t *sync, stats_t *stats, u_int32 seq)
{
  u_int32 udelta = seq - sync->max_seq;

  /* in order, with permissible gap */
  if (udelta < MAX_DROPOUT) {
    if (seq < sync->max_seq) {
      sync->cycles += VAT_SEQ_MOD;
    }
    sync->max_seq = seq;
  }
  /* possibly restart */
  else if (udelta < VAT_SEQ_MOD - MAX_MISORDER) {
    if (seq == sync->bad_seq) {
      /*
       * two sequential packets -- assume that the other side
       * restarted without telling us so just re-sync
       * (i.e., pretend this was the first packet).
       */
    }
    else {
      sync->bad_seq = (seq + 1) & (VAT_SEQ_MOD-1);
      return 0;
    }
    sync->max_seq = seq;
  }
  /* out-of-sequence packet */
  else {
    stats->misorder++;
  }

  return 1;
} /* update_seq */


/*
* Create an idlist for a single session.
* Return actual length.
*/
static int vat_write_idlist(Audio_session_t s, session_t *se)
{
  Audio_member_t m;     /* member pointer */
  vat_site_id_t *vsid;  /* current vat site identifier */
  char *id;

  se->type             = vat_idlist;
  se->d.idlist.nsids   = s->members;
  se->d.idlist.ainfo   = s->rt.audio_header;
  se->d.idlist.blksize = 0;

  vsid = &se->d.idlist.site_id;
  ALL_MEMBERS(s) {
    vsid->addr = m->src;
    id = Audio_MemberId(m);
    strcpy_pad(vsid->id, id);
    vsid = (vat_site_id_t *)ROUND((long)vsid + strlen(id) + 1, 4);
  }
  return (char *)vsid - (char *)se;
} /* vat_write_idlist */


/*
* Return control packet mbuf. 'reason' is not used.
*/
mbuf_t *vat_write_control(Audio_session_t s, int type, char *reason)
{
  mbuf_t *mb;         /* mbuf for control packet */
  session_t *se = 0;  /* mbuf as session structure */
  Audio_member_t m = s->local;

  /* 
   * Get mbuf for control packet; put data in front as there is
   * nothing else in these packets; assures proper alignment.
   */
  mb = m_get("vwc");
  if (!mb) return 0;
  mb->m_len  = 0;
  mb->m_next = 0;

  if (type & P_sdes) {
    struct sdes *sdes;
    se = mtod(mb, session_t *);
    se->type = vat_id;

    /* Insert SDES NAME into packet. */
    for (sdes = m->sdes; sdes; sdes = sdes->next) {
      if (sdes->type == SDES_NAME) {
        strncpy((char *)se->d.id, sdes->value, sdes->len);
        se->d.id[sdes->len] = '\0';
        mb->m_len = (sdes->len + 1) + sizeof(struct CtrlMsgHdr);
        break;
      }
	}
  }

  if (type & P_sdes_all) {
    se = mtod(mb, session_t *);
    se->type = vat_idlist;
    mb->m_len = vat_write_idlist(s, se);
  }

  if (type & P_bye) {
    se = mtod(mb, session_t *);
    se->type = vat_done;
    mb->m_len = sizeof(struct CtrlMsgHdr);
  }

  if (se) {
    se->flags  = 0;
    se->confid = s->sp.confid;
  }

  /* When encrypted, pad with zero bytes if needed. */
  if (s->encrypt) {
    int pad = PAD(mb->m_len, 8);
    mb->m_len += pad;
    memset(mtod_end(mb, char *), 0, pad);
  }
  return mb;
} /* vat_write_control */


/*
* Write a vat header. Return pointer to new header chain.
* 'r' determines ring where audio is from (for gateway).
*/
static mbuf_t *vat_write_data(Audio_session_t s, ring_t *r, int type,
  mbuf_t *mb)
{
  vat_hdr_t *h;
  int nsid = 0;    /* number of site entries */
  src_id_t *sid;   /* site list */
  int i;           /* loop variable */

  if (mb) {   /* audio available */
    /* count site id's */
    if (r) {
      for (i = 0; i < MAX_TALKING; i++) {
        if (r->member[i]) nsid++;
      }
      mprepend(mb, nsid*4);
    }
    mprepend(mb, 8); /* fixed header */

    h = mtod(mb, vat_hdr_t *);
    h->nsid   = nsid;
    h->flags  = s->rt.audio_header;
    if (type & P_bot) h->flags |= VATHF_NEWTS;
    h->confid = htons(s->sp.confid); 
    h->ts     = htonl(mb->m_timestamp + s->initial_ts);

    /* Add site ids. */
    if (r) {
      sid = (src_id_t *)(h+1);
      for (i = 0; i < MAX_TALKING; i++) {
        if (r->member[i]) sid[h->nsid++] = r->member[i]->src;
      }
    }
  } /* if mb */

  return mb;
} /* vat_write_data */


/*
* Parse an idlist of length 'len' from sync.  source 's' (may contain
* sources yet unknown).
*/
static void vat_parse_idlist(Audio_session_t s, Audio_member_t m,
   session_t *se, int len)
{
  int nsids;
  int i;
  vat_site_id_t *vsid;
  Audio_member_t cm;

  nsids = se->d.idlist.nsids;
  vsid  = &se->d.idlist.site_id;

  if (se->d.idlist.ainfo > 31) {
    debug(DEBUG_CNTL, "vat illegal ainfo %d (nsids=%d, type=%d)",
      se->d.idlist.ainfo, nsids, se->type);
    return;
  }
  for (i = 0; i < nsids; i++) {
    cm = member_find(s, MEMBER_CONTROL, &m->nsrc[0], vsid->addr);
    if (cm) {
      int id_len = strlen(vsid->id);
      cm->audio = vat_ad[se->d.idlist.ainfo];
      member_sdes(cm, SDES_CNAME, inet_Ntoa(vsid->addr), 0);
      member_sdes(cm, SDES_NAME,  vsid->id, id_len);
      vsid = (vat_site_id_t *)ROUND((long)vsid + sizeof(vsid->addr) + id_len+1, 4);
    }
  }
} /* vat_parse_idlist */

/*
* Process header options of control packet. Return member identifier 'm'.
*/
Audio_errno_t vat_read_control(Audio_session_t s, mbuf_t *mb,
  tsap_t *nsrc, Audio_member_t *mp)
{
  Audio_member_t m;
  session_t *se;
  int len;
  u_int32 addr = tsap_ip4(nsrc);

  m = *mp = member_find(s, MEMBER_CONTROL, nsrc, addr);
  se = mtod(mb, session_t *); 
  if (se->flags) {
    debug(DEBUG_CNTL, "vat unknown flags %x, type %d from %s",
      se->flags, se->type, Audio_MemberId(m));
  }
  switch(se->type) {
  /*
   * Vat_id is a null-terminated identifier; apparently not always
   * null-terminated, though.
   */
  case vat_id:
    len = mb->m_len - 5;      /* subtract trailing zero */
    if (se->d.id[len]) len++; /* if not zero-terminated */
    member_sdes(m, SDES_CNAME, inet_Ntoa(addr), 0);
    member_sdes(m, SDES_NAME,  se->d.id, len);
    break;
  case vat_done:
    /* If multicast, delete site. If unicast, just put it asleep. */
    if (session_ismulticast(m->s)) Audio_MemberDelete(m, "");
    else Audio_MemberClose(m, "");
    break;
  case vat_idlist:
    vat_parse_idlist(m->s, m, se, mb->m_len);
    break;
  default:
    debug(DEBUG_CNTL, "vat unknown control type %d from %s (%d bytes).",
      se->type, Audio_MemberId(m), mb->m_len);
    return Audio_EITYPE;
    break;
  } /* switch */
  return Audio_OK;
} /* vat_read_control */


/*
* Locate synchronization source. Remove header in mbuf. Return 0
* if invalid header (invalid confid).
*/
Audio_errno_t vat_read_data(Audio_session_t s, mbuf_t *mb, tsap_t *nsrc, 
  int *cc, Audio_member_t *mlist, Audio_member_t *mp)
{
  int i;
  Audio_member_t m;
  vat_hdr_t *h = mtod(mb, vat_hdr_t *);
  src_id_t *csrc = (src_id_t *)(h+1);
  int enc;
  u_int32 addr = tsap_ip4(nsrc);

  /* Check confid. */
  if (h->confid != s->sp.confid) return Audio_EICONFID;

  /* Strip header and check if still a valid packet. */
  m_adj(mb, 8 + h->nsid * 4);
  if (mb->m_len < 0) return Audio_EICSRC;

  m = *mp = member_find(s, 0, nsrc, addr);
  /* if new member, describe it */
  if (!m->id) member_sdes(m, SDES_CNAME, inet_Ntoa(addr), 0);

  /* decode audio format if it has changed (or on first packet) */
  if (h->flags & 0x60) return Audio_EIAUDIO;
  enc = h->flags & 0x1f;
  if (m->audio_header != enc) {
    m->audio_header = enc;
    if (vat_ad[enc].encoding == AE_MAX) return Audio_EIAUDIO;
    audio_out(m, &vat_ad[enc]);
  }
  /* beginning of talkspurt marker */
  m->talkspurt |= (h->flags & VATHF_NEWTS);

  /* time stamp and sequence number */
  m->sync.this.ts  = ntohl(h->ts);
  m->sync.this.seq = ntohl(h->ts);
  update_seq(&m->sync, &m->stats, m->sync.this.seq);

  /*
   * Compute expected number of packets, based on last timestamp
   * received.  Losses that are between the last packet received from
   * a talkspurt and the first in the next are not counted.  If the
   * first packet in a talkspurt is lost, packets never sent (the
   * silence period) will be counted as lost.  Thus, this estimate is
   * likely to work reasonably well only with low losses.  Reordering
   * is handled correctly.
   */
  if (m->talkspurt) {
    m->stats.expected++;
  }
  else {
    m->stats.expected += 
       ((int)(m->sync.this.ts - m->sync.prior.ts) / 
        (int)audio_bytes_to_samples(&vat_ad[m->audio_header], mb->m_len));
  }

  /* CSRC list; include SSRC for convenience */
  *cc = h->nsid & 0x0f;
  for (i = 0; i < *cc; i++) {
    mlist[i+1] = member_find(s, 0, nsrc, csrc[i]);
  }
  /* include SSRC for convenience */
  mlist[0] = m;
  *cc += 1;
  return Audio_OK;
} /* vat_read_data */

/*
* Can't reliably determine number of lost packets without a sequence
* number, but the estimate should be reasonably good.
*/
static u_int32 vat_packets_exp(Audio_member_t m)
{
  return m->stats.expected;
} /* vat_packets_exp */


/*
* Change outgoing audio encoding.
* Return -1 if no match.
*/
int vat_change_audio(rt_t *rt, audio_descr_t *ad)
{
  int i;

  /* search for matching audio descriptor for outgoing audio */
  for (i = 0; i < MAX_PT; i++) {
    if (audio_cmp(ad, &vat_ad[i])) {
      rt->audio_header = i;
      break;
    }
  }
  if (i == MAX_PT) return -1;
  return 0;
} /* vat_change_audio */

/*
* Initialize session descriptor for this protocol; set up table that
* describes all audio types known to the protocol.  Must be called
* once if incoming vat packets are to be handled, possibly with rt ==
* 0 if no outgoing vat packets are to be generated.
*/
void vat_init(rt_t *rt)
{
  int i;

  if (rt) {
    rt->write_control = vat_write_control;
    rt->write_data    = vat_write_data;
    rt->packets_exp   = vat_packets_exp;
    rt->change_audio  = vat_change_audio;
  }

  /*
   * Fill in audio_descr values on first call, but only those
   * that we understand.
   */
  if (vat_ad[0].sample_rate == 0) {
    for (i = 0; i < MAX_PT; i++) {
      vat_ad[i].encoding    = AE_MAX;
      vat_ad[i].sample_rate = 8000;
      vat_ad[i].channels    = 1;
      vat_cpp[i]            = 160;
    }
    vat_ad[VAT_AUDF_MULAW8].encoding = AE_PCMU;
    vat_ad[VAT_AUDF_CELP].encoding   = AE_1016;
    vat_ad[VAT_AUDF_GSM].encoding    = AE_GSM;
    vat_ad[VAT_AUDF_IDVI].encoding   = AE_DVI4;
    vat_ad[VAT_AUDF_LPC1].encoding   = AE_LPC;
    vat_ad[VAT_AUDF_LPC4].encoding   = AE_LPC;
  }
} /* vat_init */
