/***********************************************************************************
 *
 *  event handler for new incoming SIP messages
 *
 * Copyright 1997 by Stefan Hoffmann
 * GMD Fokus 
 *
 ***********************************************************************************/



#include <netinet/in.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include "debug.h"
#include "list.h"
#include "misc.h"
#include "notify.h"
#include "reply_codes.h"
#include "response.h"
#include "tcp.h"
#include "udp.h"



extern int                 tcp_server_socket, udp_server_socket;
extern inv_list_t          *invitation_list;



/* socket_event **********************************************************************
 *
 *  react to readable TCP or UDP socket 
 *
 *  invoked if a TCP or UDP socket becomes readable
 *  checks if incoming message is a request or response
 *  and handles it appropriately 
 *
 *  in:
 *  - client: - pointer to invitation_list entry of existing TCP connection 
 *            - NULL, if UDP or new TCP request
 *  -fd: socket at which event occurs
 *
 ***********************************************************************************/
Notify_value socket_event (Notify_client client, int fd)
{
  static char              func[] = "socket_event";
  inv_list_t               *inv;
  struct sockaddr_in       remoteAddr;
  int                      remoteAddrLen = sizeof (remoteAddr);
  char                     addr[17]; /* address from which message was sent if UDP message */
  int                      port; /* port from which message was sent if UDP message */
  int                      transport; /* transport protocol, SOCK_STREAM or SOCK_DGRAM */
  char                     *msg; /* pointer to incoming message */
  char                     method[MAX_METHOD_LENGTH];
  char                     call_id[MAX_CALL_ID_LENGTH];
  char                     user[MAX_USER_NAME_LENGTH]; /* user specified in first SIP line */
  char                     sip_version[10]; 
  char                     *p;
  int                      new_tcp_connection = FALSE; /* flag */
  

  DEBUG (2, syslog (LOG_DEBUG, "%s",func);)
    
  inv = (inv_list_t *)client;
  
  /***********************************************************************************
   * check whether message was sent via TCP or UDP and read message
   ***********************************************************************************/
  if (fd == udp_server_socket) {
    /* UDP */
    transport = SOCK_DGRAM;
    if ((msg = UDP_read(fd, &port, (char *)&addr)) == NULL) 
    return -1;
  }
  else {
    /* TCP */
    transport = SOCK_STREAM;
    if (fd == tcp_server_socket) {
      if ((fd = accept (fd, (struct sockaddr*) &remoteAddr, &remoteAddrLen)) == -1) 
        return -1;
      new_tcp_connection = TRUE;
    }
    if ((msg = TCP_read(fd, inv)) == NULL) { 
      if (new_tcp_connection) 
        close(fd);
      return (-1);
    }
  }
      
  /***********************************************************************************
   * check whether it's a request or response 
   ***********************************************************************************/
  if ((str_search(msg, SIP_VERSION)) == msg) {
    /* response */
    process_response(msg, addr);
  }
  else {
    /* request */
    /* split and check the 1st request line */ 
    p = msg;
    if ((p = get_field (p, method, sizeof (method) -1)) == NULL) {
      DEBUG (1, syslog (LOG_ERR, "%s: get_field: method", func);)
      free(msg);
      if (new_tcp_connection)
        close(fd);
      return -1;
    }
 
    if ((p = get_field (p, user, sizeof (user) -1)) == NULL) {
      DEBUG (1, syslog (LOG_ERR, "%s: get_field: user", func);)
      free(msg);
      if (new_tcp_connection)
        close(fd);
      return -1;
    }
 
    if ((p = get_field (p, sip_version, sizeof (sip_version) -1)) == NULL) {
      DEBUG (1, syslog (LOG_ERR, "%s: get_field: sip_version", func);)
      free(msg);
      if (new_tcp_connection)
        close(fd);
      return -1;
    }

    /***********************************************************************************
     * get call-id
     ***********************************************************************************/
    if ((p = search_SIP_header(msg, "\nCall-id:","\ni:")) != NULL) {
      if ((p = get_field (p, call_id, sizeof(call_id)-1)) == NULL) {
        free(msg);
        if (new_tcp_connection) 
          close(fd);
        return -1;
      }
    } else {
      /* bad request */
      free(msg);
      if (new_tcp_connection)
        close(fd);
      return -1;
    }
    
    /***********************************************************************************
     * check SIP version and method 
     ***********************************************************************************/
    if (str_cmp(sip_version, SIP_VERSION) || !str_cmp(method, "OPTIONS")) {
      if (!inv_list_add()) {
        strcpy(invitation_list->call_id, call_id);
        strcpy(invitation_list->request, msg);
        invitation_list->transport_to_client=transport;
        if (transport == SOCK_STREAM) 
        invitation_list->socket_to_client=fd;
        else { 
          strcpy(invitation_list->addr_of_client, addr);
          invitation_list->port_to_client=port;
        }
      }
      free(msg);
      reply(invitation_list, NOT_IMPLEMENTED, "Not Implemented");
      return -1;
    }
 
    /* new or retransmitted request ??? */
    if ((inv = inv_list_search(call_id)) != NULL) {
      /***********************************************************************************
       * retransmitted request
       * if a reply was already sent, retransmit the response to the client 
       ***********************************************************************************/
      if (new_tcp_connection)
        close(fd);
      /* retransmit reply */
      check_response_priority(inv, inv->response, inv->response_priority);
      free(msg);
    }
    else {
      /***********************************************************************************
       * new request
       * add to invitation list 
       ***********************************************************************************/
      if (inv_list_add()) {
        free(msg);
        return(-1);
      }
     
      /* fill invitation list entry */
      strcpy(invitation_list->call_id, call_id);
      if (strstr(user, "://") != NULL) 
        strcpy(invitation_list->user, strstr(user,"//")+2);
      else 
        strcpy(invitation_list->user, user);
      strcpy(invitation_list->request, msg);
      invitation_list->transport_to_client=transport;
      if (transport == SOCK_STREAM) {
        /* TCP connection to client */
        invitation_list->socket_to_client=fd;
        /* install event handler for (accepted) socket */
        notify_set_input_func((Notify_client)invitation_list, (Notify_func_input) socket_event, invitation_list->socket_to_client);    
      }
      else { 
        /* request was received via UDP */
        invitation_list->port_to_client=port;
        strcpy(invitation_list->addr_of_client, addr);
      }
      free(msg);
      
      /***********************************************************************************
       * notify client that the request was received
       ***********************************************************************************/  
      reply(invitation_list, TRYING, "Trying");
     
      process_request(invitation_list);
    }
  }
  return;
} /* socket_event */


