#ifndef vic_ac_h
#define vic_ac_h

#include "Tcl.h"

#define WORST_RECEIVER 0
#define CONGESTION_RATIO 1
#define LOSS_DELAY_BASED_ADJUSTMENT 3
#define LOSS_BASED_ADJUSTMENT 2
#define DIRECT_ADJUSTMENT 5

#define CONTROL                    10
#define ALGORITHM                  11
#define ALPHA                      12
#define MAX_BANDWIDTH              13
#define MIN_BANDWIDTH              14
#define MULTIPLICATIVE_DECREASE    15
#define ADDITATIVE_DECREASE        16
#define LOSS_LOWER_THRESHOLD       17
#define LOSS_UPPER_THRESHOLD       18
#define CONGESTED_RECV_FRACTION    19
#define LOADED_RECV_FRACTION       20

#define UPDATE_INTERVAL  5000000.0  /* the size of an update interval in usec */

typedef struct {
  double filtered_loss;     /* linear filtered loss */
  double loss;              /* last reported loss  0<= loss <=1 */
  int ssrc;                 /* actual receiver ssrc */

  double current_Q;         /* last measured queuing delay */
  double max_Q;             /* last seen queueing delay before a loss */
  double rtt;               /* last measured round trip time */
  double min_rtt;           /* smalles measured round trip time, isconsidered as the propagation delay */
  u_int32_t time;                 /* update time of this member in usecs */
  double TCP_Rate;          /* TCP rate for loss and min_rtt */
  double bbw;               /* bottleneck bandwdith for thos member */
} mem_t; 


typedef enum {UNLOADED, LOADED, CONGESTED} network_state_t;     
class AC : public TclObject {
public: 
  AC();
  virtual ~AC();
  /* interface to the TCL Interpreter */
  int command(int argc, const char*const* argv);
  /* initialize the needed information structure for the application
   * controller
   */
  void ac_init(u_int32_t Ssrc);

  /* calculate the appropriate bandwidth to use */
  int ac_qos_control(int Ssrc, double Loss, double rtt,int bandwidth);

  /* min_bandwidth is set by default to 0. In case of using some kind of
   * guranteed reservation the min_bandwidth value can be accordungly changed
   */
  void ac_update(int tag, int parameter);

  /* return the max bandwidth value of the controller */
  int ac_max_bandwidth(); 

  /*
   * add an entry to the table
   */
  mem_t* ac_add_entry(u_int32_t Ssrc);

  /* delete the entry made for the stream with the ssrc value */
  void ac_delete_entry(u_int32_t Ssrc);

  /* update the bbw vale */
  void ac_update_bbw(int Ssrc,double bbw);

private:
  u_int32_t ssrc;                      /* ssrc of controlled sender */
  /*
  int (*analyse)();
  int (*adjust)();
  */
  Tcl_HashTable member_table;        /* hash table for members */
  int members;                       /* # of members including the sender */

  /* adjustment specific variables */
  int control;
  double min_bandwidth;                /* minimal bw value under which no control
                                     * should be used */
  double max_bandwidth;                /* maximal bw value under which no control
                                     * should be used */
  int allowed_bandwidth;             /* calculated bw to use */
  double alpha;                      /* used for the filter */
  double multiplicative_decrease;    /* value with which to decrease the bw */
  double additive_increase;          /* initial value with which to increase 
                                        bw */
  double aif;                        /* current additive_increase factor*/
  double loss_lower_threshold;
  double loss_upper_threshold;
  double congested_recv_fraction;
  double loaded_recv_fraction;
  int algorithm;                    /* WORST_RECEIVER or CONGESTION_RATIO */
  double loss;                      /* the heighest loss seen in RR*/
  u_int32_t ssrc_loss;                /* ssrc of the source reporting the
                                     * heighets loss */
  int time;                         /* time of last update interval in usec*/ 

  /*
   * get the time in usecs
   */
  u_int32_t ac_time();
  /*
   * convenience function for searching an entry in the hash table
   * returns pointer to member or NULL 
   */
  mem_t* ac_find_entry(u_int32_t Ssrc);
  /*
   * This function analyses the information contained in the RTP database 
   * and returns the actual network state.
   * The network state is set according to the maximum loss rate of all
   * receivers.
   */
  network_state_t worst_receiver();
  
  /* 
   * Count the number of receivers that are congested, loaded and unloaded
   * and determine network state 
   */
  network_state_t congestion_ratio();
  
  /*
   * low-pass filter to compute the short time statistics                 
   * the smaller a, the greater is the influence of the old value 
   */
  float ac_filter(float oldval, float newval, float a);

  /*
   * This function interpretes the network state and adjusts the bandwidth
   * of the video agent accordingly
   */
  int adjust(network_state_t state,int bandwidth);
  
  /* the direct_adjustment algorithm reduces the sending bandwidth directly 
   * to the level reported in the receiver reports. the reduction is then
   * maintained untill a new RR arrives from the same receiver indicating that
   * the loss is beneath the lower loss threshold.
   * In the case that after reducing the rate another RR message was received
   * from another receiver with a heigher loss ratio then the rate is only
   * reduced by the difference of the losses reported in this RR and the one
   * that caused the original loss. Afterwards we have to wait for another RR
   * from this receiver to start increasing the rate again.
   */ 
  int direct_adjustment(int ssrc, double loss, int bandwidth);

  /*
   * update the entries calculation for this member
   */
  void ac_update_entry(mem_t* m );

  /*
   * calculate for each member the bandwdith it could have been getting
   * under the condition of current bandwdith and the measured delay and loss
   */
  double ac_get_LDBA_bw(mem_t* m,double current_bandwidth); 

  /*
   * find the smallest possible bandwdith to use 
   */
  double ac_get_min_entry(double current_bandwidth);
  
  int ac_min_bbw();
};  


#endif
