/* 
 * tclUnixUdp.c --
 *
 *	This file contains Unix-specific udp socket related code.
 *
 */

#include "tcl.h"
#include "tclInt.h"
#include "tclPort.h"

/*
#if defined(linux)
#define ISMULTICAST(x) ( (x & 0xE0) == 0xE0 )
#endif
#if defined(solaris)
#define ISMULTICAST(x) ( (x & 0xE0) == 0xE0 )
#endif
*/

#if !defined(linux)
#define ISMULTICAST(x) ( (x & 0xF0000000) == 0xE0000000 )
#else
#define ISMULTICAST(x) ( (x & 0x000000F0) == 0x000000E0 )
#endif

/*
 * This structure describes per-instance state of a udp based channel.
 */

typedef struct UdpState {
    int flags;				/* ORed combination of the
                                         * bitfields defined below. */
    Tcl_File sock;			/* The socket itself. */
    int doread;				/* return data from UdpInput? */
    int member;				/* member of multicast group? */
    struct sockaddr_in to;              /* Where is this data going to? */
    struct sockaddr_in from;            /* Where did last data come from? */
} UdpState;

/*
 * The following defines how much buffer space the kernel should maintain
 * for a socket.  This is also defined in tclUnixChan.c!
 */
#define SOCKET_BUFSIZE	4096

/*
 * These bits may be ORed together into the "flags" field of a UdpState
 * structures.  This is also defined in tclUnixChan.c!
 */
#define TCP_ASYNC_SOCKET	(1<<0)

static int		UdpBlockMode _ANSI_ARGS_((ClientData data,
		            Tcl_File inFile, Tcl_File outFile, int mode));
static UdpState *	CreateUdpSocket _ANSI_ARGS_((Tcl_Interp *interp,
			    int port, char *host, int server,
			    char *myaddr, int myport));
static int		UdpInput _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File infile, char *buf, int toRead,
        		    int *errorCode));
static int		UdpOutput _ANSI_ARGS_((ClientData instanceData,
		            Tcl_File outFile, char *buf, int toWrite,
		            int *errorCode));
static int		UdpSetOption _ANSI_ARGS_((ClientData instanceData,
			    Tcl_Interp *interp, char *optionName,
			    char *newValue));
static int		UdpGetOption _ANSI_ARGS_((ClientData instanceData,
		            char *buf, Tcl_DString *dsPtr));
static int		UdpClose _ANSI_ARGS_((ClientData instanceData,
			    Tcl_Interp *interp, Tcl_File inFile,
                            Tcl_File outFile));

/*
 * This structure describes the channel type structure for UDP socket
 * based IO:
 */

static Tcl_ChannelType udpChannelType = {
    "udp",				/* Type name. */
    UdpBlockMode,			/* Set blocking/nonblocking mode.*/
    UdpClose,				/* Close proc. */
    UdpInput,				/* Input proc. */
    UdpOutput,				/* Output proc. */
    NULL,				/* Seek proc. */
    UdpSetOption,			/* Set option proc. */
    UdpGetOption,			/* Get option proc. */
};


/*
 *----------------------------------------------------------------------
 *
 * UdpCreateSocketAddress --
 *
 *	This function initializes a sockaddr structure for a host and port.
 *
 * Results:
 *	1 if the host was valid, 0 if the host could not be converted to
 *	an IP address.
 *
 * Side effects:
 *	Fills in the *sockaddrPtr structure.
 *
 *----------------------------------------------------------------------
 */

int
UdpCreateSocketAddress(sockaddrPtr, host, port)
    struct sockaddr_in *sockaddrPtr;	/* Socket address */
    char *host;				/* Host.  NULL implies INADDR_ANY */
    int port;				/* Port number */
{
    struct hostent *hostent;		/* Host database entry */
    struct in_addr addr;		/* For 64/32 bit madness */

    (void) memset((char *) sockaddrPtr, '\0', sizeof(struct sockaddr_in));
    sockaddrPtr->sin_family = AF_INET;
    sockaddrPtr->sin_port = htons((unsigned short) (port & 0xFFFF));
    if (host == NULL) {
	addr.s_addr = INADDR_ANY;
    } else {
	hostent = gethostbyname(host);
	if (hostent != NULL) {
	    memcpy((char *) &addr,
		(char *) hostent->h_addr_list[0], (size_t) hostent->h_length);
	} else {
	    addr.s_addr = inet_addr(host);
	    if ( (addr.s_addr == (unsigned long)-1) &&
		 (strcmp(host, "255.255.255.255") != 0) ) {
		return 0;	/* error */
	    }
	}
    }
    /*
     * There is a rumor that this assignment may require care on
     * some 64 bit machines.
     */

    sockaddrPtr->sin_addr.s_addr = addr.s_addr;
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * is_addr_multi --
 *
 *      Determine whether an IP address is multicast or not.  We don't
 *	bother to check the validity of the address (it must be in
 *	dotted decimal to succeed), instead we let inet_addr do this.
 *
 * Results:
 *      1 if multicast, 0 otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int
is_addr_multi(nam)
    char *nam;
{
    unsigned long addr;

    addr = inet_addr(nam);
    if (addr != -1) {
        return ISMULTICAST(addr);
    }

    return 0;
} 

/*
 *----------------------------------------------------------------------
 *
 * UdpClose --
 *
 *	Close the socket.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the socket.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
UdpClose(instanceData, interp, inFile, outFile)
    ClientData instanceData;	/* The socket to close. */
    Tcl_Interp *interp;		/* For error reporting - unused. */
    Tcl_File inFile, outFile;	/* Unused. */
{
    UdpState *statePtr;
    Tcl_File sockFile;
    int sock;
    int errorCode = 0;

    statePtr = (UdpState *) instanceData;
    sockFile = statePtr->sock;
    sock = (int) Tcl_GetFileInfo(sockFile, NULL);
    
    /*
     * Delete a file handler that may be active for this socket.
     * Channel handlers are already deleted in the generic IO close
     * code which called this function.
     */
    
    Tcl_DeleteFileHandler(sockFile);

    ckfree((char *) statePtr);
    
    /*
     * We assume that inFile==outFile==sockFile and so
     * we only clean up sockFile.
     */

    Tcl_FreeFile(sockFile);

    if (close(sock) < 0) {
	errorCode = errno;
    }

    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * UdpBlockMode --
 *
 *	Set blocking or non-blocking mode on a UDP channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
UdpBlockMode(instanceData, inFile, outFile, mode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File inFile, outFile;		/* Input, output for channel. */
    int mode;				/* The mode to set. */
{
    int curStatus;
    int sock;
    UdpState *statePtr;
    
    statePtr = (UdpState *) instanceData;
    if (mode == TCL_MODE_BLOCKING) {
        statePtr->flags &= (~(TCP_ASYNC_SOCKET));
    } else {
        statePtr->flags |= TCP_ASYNC_SOCKET;
    }
    if (inFile != NULL) {
        sock = (int) Tcl_GetFileInfo(inFile, NULL);
        curStatus = fcntl(sock, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(sock, F_SETFL, curStatus) < 0) {
            return errno;
        }
    }
    if (outFile != NULL) {
        sock = (int) Tcl_GetFileInfo(outFile, NULL);
        curStatus = fcntl(sock, F_GETFL);
        if (mode == TCL_MODE_BLOCKING) {
            curStatus &= (~(O_NONBLOCK));
        } else {
            curStatus |= O_NONBLOCK;
        }
        if (fcntl(sock, F_SETFL, curStatus) < 0) {
            return errno;
        }
    }

    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * UdpInput --
 *
 *	Reads input from the UDP IO channel into the buffer given.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
UdpInput(instanceData, inFile, buf, bufSize, errorCode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File inFile;			/* Input device for channel. */
    char *buf;				/* Where to store data read. */
    int bufSize;			/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    UdpState *statePtr;			/* The state of the socket. */
    int sock;				/* The OS handle. */
    int bytesRead;			/* How many bytes were read?  */
    int len;				/* length of sockaddr (why?)  */

    statePtr = (UdpState *) instanceData;

    /*
     * The caller of this function is looking for a stream oriented
     * system, so it keeps calling the function until no bytes are
     * returned, and then appends all the characters together.  This
     * is not what we want from UDP, so we fake it by returning a
     * blank every other call.  whenever the doread variable is 1 do
     * a normal read, otherwise just return 0.
     */
    if (statePtr->doread == 0) {
        statePtr->doread = 1;  /* next time we want to behave normally */
        *errorCode = EAGAIN;    /* pretend that we would block */
        return 0;
    }

    *errorCode = 0;
    errno = 0;
    sock = (int) Tcl_GetFileInfo(inFile, NULL);
    
    if (bufSize == 0) {
        return 0;
    }

    len = sizeof(statePtr->from);
    bytesRead = recvfrom(sock, buf, bufSize, 0,
		         (struct sockaddr *) &(statePtr->from), &len);

    /* we don't want to return anything next time */
    if (bytesRead > 0) { buf[bytesRead] = '\0'; statePtr->doread = 0; }

    if (bytesRead > -1) {
        return bytesRead;
    }

    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * UdpOutput--
 *
 *	Writes the given output on the UDP IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

static int
UdpOutput(instanceData, outFile, buf, toWrite, errorCode)
    ClientData instanceData;		/* Socket state. */
    Tcl_File outFile;			/* Output device for channel. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    UdpState *statePtr;
    int written;
    int sock;
    int len;

    *errorCode = 0;
    errno = 0;
    sock = (int) Tcl_GetFileInfo(outFile, NULL);
    statePtr = (UdpState *) instanceData;

    len = sizeof(statePtr->to);
    written = sendto(sock, buf, toWrite, 0,
		     (struct sockaddr *)&(statePtr->to), len);
    if (written > -1) {
        return written;
    }
    *errorCode = errno;
    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * JoinMulticastGroup --
 *
 *      This function attempts to join a multicast group address.
 *
 * Results:
 *      0 on success, < 0 on failure.
 *
 * Side effects:
 *      On success this machine will now listen to the multicast
 *	address.
 *
 *----------------------------------------------------------------------
 */
static int
JoinMulticastGroup(sock, addr)
    int sock;
    char *addr;
{
    int status;
    struct ip_mreq imr;

    /* tell IGMP we are interested */
    imr.imr_multiaddr.s_addr = inet_addr(addr);
    imr.imr_interface.s_addr = htonl(INADDR_ANY);
    status = setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &imr,
                        sizeof(struct ip_mreq));

    return status;
}

/*
 *----------------------------------------------------------------------
 *
 * DropMulticastGroup --
 *
 *      This function attempts to leave a multicast group.
 *
 * Results:
 *      0 on success, < 0 on failure.
 *
 * Side effects:
 *      On success this machine will quit receiving packets destined
 *	for the multicast address.
 *
 *----------------------------------------------------------------------
 */
static int
DropMulticastGroup(sock, addr)
    int sock;
    char *addr;
{
    int status;
    struct ip_mreq imr;

    /* tell IGMP we are no longer interested */
    imr.imr_multiaddr.s_addr = inet_addr(addr);
    imr.imr_interface.s_addr = htonl(INADDR_ANY);
    status = setsockopt(sock, IPPROTO_IP, IP_DROP_MEMBERSHIP, (char *) &imr,
                        sizeof(struct ip_mreq));

    return status;
}

/*
 *----------------------------------------------------------------------
 *
 * CreateUdpSocket --
 *
 *	This function opens a new socket in client or server mode
 *	and initializes the UdpState structure.
 *
 * Results:
 *	Returns a new UdpState, or NULL with an error interp.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static UdpState *
CreateUdpSocket(interp, port, host, server, myaddr, myport)
    Tcl_Interp *interp;		/* For error reporting; can be NULL. */
    int port;			/* Port number to open. */
    char *host;			/* Name of host on which to open port.
				 * NULL implies INADDR_ANY */
    int server;			/* 1 if socket should be a server socket,
				 * else 0 for a client socket. */
    char *myaddr;		/* Optional client-side address */
    int myport;			/* Optional client-side port */
{
    int status, sock;
    struct sockaddr_in sockaddr;	/* socket address */
    struct sockaddr_in mysockaddr;	/* Socket address for client */
    UdpState *statePtr;

    if (! UdpCreateSocketAddress(&sockaddr, host, port)) {
	goto addressError;
    }
    if ((myaddr != NULL || myport != 0) &&
	    ! UdpCreateSocketAddress(&mysockaddr, myaddr, myport)) {
	goto addressError;
    } else {
	/* set up some defaults for a client who doesn't specify anything */
        (void) memset((char *) &mysockaddr, '\0', sizeof(struct sockaddr_in));
	mysockaddr.sin_family = AF_INET;
	mysockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	mysockaddr.sin_port = htons(0);
    }

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
	goto addressError;
    }

    /*
     * Set kernel space buffering
     */
    TclSockMinimumBuffers(sock, SOCKET_BUFSIZE);

    status = 1;
    (void) setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &status,
		      sizeof(status));

    if (server) {
	/*
	 * Set up to reuse server addresses automatically and bind to the
	 * specified port.
	 */
    
	status = bind(sock, (struct sockaddr *) &sockaddr,
                sizeof(struct sockaddr));
    } else if (is_addr_multi(host)) {
	/* must always bind these sockets */
	UdpCreateSocketAddress(&mysockaddr, (char *) NULL, port);
	status = bind(sock, (struct sockaddr *) &mysockaddr,
	         sizeof(struct sockaddr));
    } else {
	status = bind(sock, (struct sockaddr *) &mysockaddr,
	         sizeof(struct sockaddr));
    }

    if (status < 0) {
        if (interp != NULL) {
            Tcl_AppendResult(interp, "couldn't open udp socket: ",
                    Tcl_PosixError(interp), (char *) NULL);
        }
        close(sock);
        return NULL;
    }

    /*
     * Allocate a new UdpState for this socket.
     */
    statePtr = (UdpState *) ckalloc((unsigned) sizeof(UdpState));
    statePtr->flags = 0;
    statePtr->doread = 1;
    statePtr->member = 0;
    statePtr->sock = Tcl_GetFile((ClientData) sock, TCL_UNIX_FD);

    if (!server) {
        /*
         * If we aren't working with a server, give the client somewhere to
         * send his information to.  Remember we're using sendto for UDP!
         */
        memcpy((void *)&(statePtr->to), (void *)&sockaddr, sizeof(sockaddr));
    } else {
	/*
	 * Clear the "to" field if this is a server, because we don't want to
	 * have a "to" until someone has actually sent us something, and the
	 * programmer has set it.  Let the call to recvfrom set the "from"
	 * fields.
	 */
        (void)memset((void *)&(statePtr->to), '\0', sizeof(sockaddr));
    }

    return statePtr;

  addressError:
    if (interp != NULL) {
	Tcl_AppendResult(interp, "couldn't open udp socket: ",
		Tcl_PosixError(interp), (char *) NULL);
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenUdpClient --
 *
 *	Opens a UDP client socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.  An error message is returned
 *	in the interpreter on failure.
 *
 * Side effects:
 *	Opens a client socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenUdpClient(interp, port, host, myaddr, myport)
    Tcl_Interp *interp;			/* For error reporting; can be NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Host on which to open port. */
    char *myaddr;			/* Client-side address */
    int myport;				/* Client-side port */
{
    Tcl_Channel chan;
    UdpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateUdpSocket(interp, port, host, 0, myaddr, myport);
    if (statePtr == NULL) {
	return NULL;
    }

    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    chan = Tcl_CreateChannel(&udpChannelType, channelName, statePtr->sock,
	    statePtr->sock, (ClientData) statePtr);

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenUdpServer --
 *
 *	Opens a UDP server socket and creates a channel around it.
 *
 * Results:
 *	The channel or NULL if failed.
 *
 * Side effects:
 *	Opens a server socket and creates a new channel.
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenUdpServer(interp, port, host)
    Tcl_Interp *interp;			/* For error reporting - may be
                                         * NULL. */
    int port;				/* Port number to open. */
    char *host;				/* Name of local host. */
{
    Tcl_Channel chan;
    UdpState *statePtr;
    char channelName[20];

    /*
     * Create a new client socket and wrap it in a channel.
     */

    statePtr = CreateUdpSocket(interp, port, host, 1, NULL, 0);
    if (statePtr == NULL) {
	return NULL;
    }

    sprintf(channelName, "sock%d",
	    (int) Tcl_GetFileInfo(statePtr->sock, NULL));

    /*
     * Creates a new entry in the hash table for a Tcl_Channel
     * record.
     */
    /*
     * Note that I have changed this from the Tcp to take an infile and
     * an outfile.
     */
    chan = Tcl_CreateChannel(&udpChannelType, channelName, statePtr->sock,
            statePtr->sock, (ClientData) statePtr);

    return chan;
}

/*
 *----------------------------------------------------------------------
 *
 * UdpSetOption --
 *
 *	This function sets the various option fields of the UDP socket.
 *
 * Results:
 *	A standard Tcl result.  Also sets the interp->result on error if
 *	interp is not NULL.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */
int
UdpSetOption (instanceData, interp, optionName, newValue)
    ClientData instanceData;
    Tcl_Interp *interp;
    char *optionName;
    char *newValue;
{
    UdpState *statePtr;
    unsigned char ttl, loop;
    int status, sock, tmp, bcast, member, tmp_ttl;

    statePtr = (UdpState *)instanceData;

    /* set the port to which we are sending information */
    if (strcmp(optionName, "-toport") == 0) {
        statePtr->to.sin_port = htons((short)atoi(newValue));
        return TCL_OK;
    }

    /* set the address to which we are sending information */
    if (strcmp(optionName, "-toaddr") == 0) {
	sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	if (is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
	    status = DropMulticastGroup(sock, inet_ntoa(statePtr->to.sin_addr));
	    if (status < 0) {
                if (interp != (Tcl_Interp *) NULL) {
                    Tcl_AppendResult(interp, "could not leave multicast group:",
     			        Tcl_PosixError(interp), (char *) NULL);
		}
                return TCL_ERROR;
	    }
            statePtr->member = 0;
	}
        statePtr->to.sin_addr.s_addr = inet_addr(newValue);
	if (is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
	    status = JoinMulticastGroup(sock, inet_ntoa(statePtr->to.sin_addr));
	    if (status < 0) {
                if (interp != (Tcl_Interp *) NULL) {
                    Tcl_AppendResult(interp, "could not join multicast group:",
     			        Tcl_PosixError(interp), (char *) NULL);
		}
                return TCL_ERROR;
	    }
            statePtr->member = 1;
	}
        return TCL_OK;
    }

    /* if this is a multicast socket, set the ttl */
    if (strcmp(optionName, "-ttl") == 0) {
	if (!is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "valid only for multicast sockets",
     			        (char *) NULL);
            }
            return TCL_ERROR;
        }

	/* do this weird stuff because ttl must be an unsigned char, but */
	/* in order to check whether newValue is >= 0 we must change it  */
	/* to an INT, because if we change it to an unsigned char, of    */
	/* course it will be >= 0.  so make it an int first, do the      */
	/* check, and _then_ chhange it to an unsigned char...		 */
	tmp_ttl = atoi(newValue);
        if (tmp_ttl >= 0) {
            ttl = (unsigned char)atoi(newValue);
	    sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
            status = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl,
				sizeof(ttl));
	    if (status < 0) {
                if (interp != (Tcl_Interp *) NULL) {
                    Tcl_AppendResult(interp, "could not change ttl: ",
     			        Tcl_PosixError(interp), (char *) NULL);
		}
                return TCL_ERROR;
	    }
        } else {
	    if (interp != (Tcl_Interp *) NULL) {
		Tcl_AppendResult(interp, "invalid ttl", (char *) NULL);
	    }
	    return TCL_ERROR;
	}
        return TCL_OK;
    }

    /* if this is a multicast socket set whether to receive our own */
    /* packets.							    */
    if (strcmp(optionName, "-loopback") == 0) {
	if (!is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "valid only for multicast sockets",
     			        (char *) NULL);
            }
            return TCL_ERROR;
        }

        if (Tcl_GetBoolean(interp, newValue, &tmp) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	loop = (unsigned char) tmp;
        status = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_LOOP, &loop,
			    sizeof(loop));
        if (status < 0) {
	    if (interp != (Tcl_Interp *) NULL) {
	    	Tcl_AppendResult(interp, "could not change loopback: ", 
		    		 Tcl_PosixError(interp), (char *) NULL);
	    }
	    return TCL_ERROR;
	}
	return TCL_OK;
    }

    if (strcmp(optionName, "-member") == 0) {
	if (!is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
            if (interp != (Tcl_Interp *) NULL) {
                Tcl_AppendResult(interp, "valid only for multicast sockets",
     			        (char *) NULL);
            }
            return TCL_ERROR;
        }
	if (Tcl_GetBoolean(interp, newValue, &member) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (member != statePtr->member) {
	    sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	    if (member == 1) {
		status = JoinMulticastGroup(sock,
				inet_ntoa(statePtr->to.sin_addr));
		if (status < 0) {
		    if (interp != (Tcl_Interp *) NULL) {
			Tcl_AppendResult(interp, "could not join multicast 
				group:",Tcl_PosixError(interp), (char *) NULL);
		    }
		    return TCL_ERROR;
		}
	    } else {
	        status = DropMulticastGroup(sock,
				inet_ntoa(statePtr->to.sin_addr));
	        if (status < 0) {
		    if (interp != (Tcl_Interp *) NULL) {
		        Tcl_AppendResult(interp, "could not leave multicast
			      group:", Tcl_PosixError(interp), (char *) NULL);
		    }
		    return TCL_ERROR;
	        }
	    }
            statePtr->member = member;
        }
        return TCL_OK;
    }


    if (strcmp(optionName, "-broadcast") == 0) {
	if (Tcl_GetBoolean(interp, newValue, &bcast) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	status = setsockopt(sock, SOL_SOCKET, SO_BROADCAST,
			    (char *) &bcast, sizeof(bcast));
	if (status < 0) {
	    if (interp != (Tcl_Interp *) NULL) {
	    	Tcl_AppendResult(interp, "could not change broadcast: ", 
		    		 Tcl_PosixError(interp), (char *) NULL);
	    }
	    return TCL_ERROR;
	}
	return TCL_OK;
    }

    if (interp != (Tcl_Interp *) NULL) {
	if (is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
            Tcl_AppendResult(interp, "bad option \"", optionName,
                "\": should be -blocking, -buffered, -buffersize, ",
                "-translation, -broadcast, -ttl, -toport, ",
		"-loopback, -member, or -toaddr",
                (char *) NULL);
	} else {
            Tcl_AppendResult(interp, "bad option \"", optionName,
                "\": should be -blocking, -buffered, -buffersize, ",
                "-translation, -broadcast, -toport, or -toaddr",
                (char *) NULL);
	}
    }
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * UdpGetOption --
 *
 *	This function gets the various option fields of the UDP socket.
 *
 * Results:
 *	The address information is returned in interp->result as
 *	a list consisting of {type, address, port}.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
UdpGetOption (instanceData, optionName, dsPtr) 
    ClientData instanceData;
    char *optionName;
    Tcl_DString *dsPtr;
{
    char tmp[80];
    UdpState *statePtr;
    size_t len;
    unsigned char loop, bcast, ttl;
    int sock, status, siz;
  
    statePtr = (UdpState *)instanceData;

    /*
     * If the optionName is NULL it means that we want a list of all
     * options and values.
     */

    if (optionName == (char *) NULL) {
        len = 0;
    } else {
        len = strlen(optionName);
    }

    if ( (len == 0) || (strcmp(optionName, "-toport") == 0) ) {
	if (len == 0) Tcl_DStringAppendElement(dsPtr, "-toport");
        sprintf(tmp, "%hd", ntohs(statePtr->to.sin_port));
        Tcl_DStringAppendElement(dsPtr, tmp);
	if (len != 0) return TCL_OK;
    }

    if ( (len == 0) || (strcmp(optionName, "-toaddr") == 0) ) {
        if (len == 0) Tcl_DStringAppendElement(dsPtr, "-toaddr");
        Tcl_DStringAppendElement(dsPtr, inet_ntoa(statePtr->to.sin_addr));
	if (len != 0) return TCL_OK;
    }

    if ( (len == 0) || (strcmp(optionName, "-fromport") == 0) ) {
        if (len == 0) Tcl_DStringAppendElement(dsPtr, "-fromport");
        sprintf(tmp, "%hd", ntohs(statePtr->from.sin_port));
        Tcl_DStringAppendElement(dsPtr, tmp);
	if (len != 0) return TCL_OK;
    }

    if ( (len == 0) || (strcmp(optionName, "-fromaddr") == 0) ) {
        if (len == 0) Tcl_DStringAppendElement(dsPtr, "-fromaddr");
        Tcl_DStringAppendElement(dsPtr, inet_ntoa(statePtr->from.sin_addr));
	if (len != 0) return TCL_OK;
    }

    if ( (len == 0) || (strcmp(optionName, "-ttl") == 0) ) {
	if (is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
            if (len == 0) Tcl_DStringAppendElement(dsPtr, "-ttl");

	    sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	    siz = sizeof(ttl);
            status = getsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, &siz);
 	    if (status < 0) {
		Tcl_DStringInit(dsPtr);
	        Tcl_DStringAppend(dsPtr, "could not retrieve ttl: ", -1);
	        Tcl_DStringAppend(dsPtr, Tcl_ErrnoMsg(errno), -1);
	        return TCL_OK;
	    }
            sprintf(tmp, "%u", ttl);
            Tcl_DStringAppendElement(dsPtr, tmp);
	    if (len != 0) return TCL_OK;
	} else if (len != 0) {
            Tcl_DStringAppendElement(dsPtr, "valid only for multicast sockets");
	}
    }

    if ( (len == 0) || (strcmp(optionName, "-loopback") == 0) ) {
	if (is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
	    if (len == 0) Tcl_DStringAppendElement(dsPtr, "-loopback");
	    sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	    siz = sizeof(loop);
            status = getsockopt(sock, IPPROTO_IP, IP_MULTICAST_LOOP,
	 		        &loop, &siz);
 	    if (status < 0) {
		Tcl_DStringInit(dsPtr);
	        Tcl_DStringAppend(dsPtr, "could not retrieve loopback: ", -1);
	        Tcl_DStringAppend(dsPtr, Tcl_ErrnoMsg(errno), -1);
	        return TCL_OK;
	    }
	    sprintf(tmp, "%d", loop);
            Tcl_DStringAppendElement(dsPtr, tmp);
	    if (len != 0) return TCL_OK;
	} else if (len != 0) {
            Tcl_DStringAppendElement(dsPtr, "valid only for multicast sockets");
	}
    }

    if ( (len == 0) || (strcmp(optionName, "-member") == 0) ) {
	if (is_addr_multi(inet_ntoa(statePtr->to.sin_addr))) {
	    if (len == 0) Tcl_DStringAppendElement(dsPtr, "-member");
	    sprintf(tmp, "%d", statePtr->member);
            Tcl_DStringAppendElement(dsPtr, tmp);
  	    if (len != 0) return TCL_OK;
	} else if (len != 0) {
            Tcl_DStringAppendElement(dsPtr, "valid only for multicast sockets");
	}
    }

    if ( (len == 0) || (strcmp(optionName, "-broadcast") == 0) ) {
	if (len == 0) Tcl_DStringAppendElement(dsPtr, "-broadcast");
	sock = (int) Tcl_GetFileInfo(statePtr->sock, NULL);
	siz = sizeof(bcast);
        status = getsockopt(sock, SOL_SOCKET, SO_BROADCAST,
	 		        &bcast, &siz);
 	if (status < 0) {
	    Tcl_DStringInit(dsPtr);
	    Tcl_DStringAppend(dsPtr, "could not retrieve broadcast: ", -1);
	    Tcl_DStringAppend(dsPtr, Tcl_ErrnoMsg(errno), -1);
	    return TCL_OK;
	}
	sprintf(tmp, "%d", bcast);
        Tcl_DStringAppendElement(dsPtr, tmp);
	if (len != 0) return TCL_OK;
    }

    if (len == 0) return TCL_OK;

    return TCL_ERROR;
}

